# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
import tkinter as tk
from enum import IntEnum
from pathlib import Path
from tkinter import ttk
from typing import Any, Dict, List, Tuple

from discopop_library.CodeGenerator.classes.UnpackedSuggestion import UnpackedSuggestion
from DEPRECATED_discopop_wizard.classes.CodePreview import CodePreviewContentBuffer


class PragmaType(IntEnum):
    PRAGMA = 1
    REGION = 2


class Suggestion(UnpackedSuggestion):
    def __init__(self, wizard, type_str: str, values: Dict[str, Any]):
        super().__init__(type_str, values)
        self.wizard = wizard

    def show_code_section(self, parent_frame: tk.Frame, execution_configuration):
        # close elements of parent_frame
        for c in parent_frame.winfo_children():
            c.destroy()

        # configure parent_frame
        parent_frame.rowconfigure(0, weight=1)
        parent_frame.columnconfigure(0, weight=1)

        # create content frame and scroll bars
        source_code = tk.Text(parent_frame, wrap=tk.NONE)
        source_code.grid(row=0, column=0, sticky="nsew")

        # create a Scrollbar and associate it with the content frame
        y_scrollbar = ttk.Scrollbar(parent_frame, command=source_code.yview)
        y_scrollbar.grid(row=0, column=1, sticky="nsew")
        source_code["yscrollcommand"] = y_scrollbar.set

        x_scrollbar = ttk.Scrollbar(parent_frame, orient="horizontal", command=source_code.xview)
        x_scrollbar.grid(row=1, column=0, columnspan=2, sticky="nsew")
        source_code["xscrollcommand"] = x_scrollbar.set

        # load file mapping from project path
        file_mapping: Dict[int, Path] = dict()
        with open(
            os.path.join(execution_configuration.value_dict["working_copy_path"], "FileMapping.txt"),
            "r",
        ) as f:
            for line in f.readlines():
                line = line.replace("\n", "")
                split_line = line.split("\t")
                id = int(split_line[0])
                path = split_line[1]
                file_mapping[id] = Path(path)

        # create CodePreview object
        code_preview = CodePreviewContentBuffer(self.wizard, self.file_id, file_mapping[self.file_id])

        # get and insert pragmas
        pragmas = self.get_pragmas()
        for pragma in pragmas:
            compile_check_command = (
                "cd "
                + execution_configuration.value_dict["working_copy_path"]
                + " && make "
                + execution_configuration.value_dict["make_flags"]
                + " "
                + execution_configuration.value_dict["make_target"]
            )
            successful = code_preview.add_pragma(
                file_mapping,
                pragma,
                [],
                skip_compilation_check=True if self.wizard.settings.code_preview_disable_compile_check == 1 else False,
                compile_check_command=compile_check_command,
            )
            # if the addition resulted in a non-compilable file, add the pragma as a comment
            if not successful:
                # print error codes
                self.wizard.console.print(code_preview.compile_result_buffer)
                code_preview.add_pragma(
                    file_mapping,
                    pragma,
                    [],
                    add_as_comment=True,
                    skip_compilation_check=True,
                    compile_check_command=compile_check_command,
                )

        # show CodePreview
        code_preview.show_in(source_code)

        # show targeted code section
        code_preview.jump_to_first_modification(source_code)

        # disable source code text widget to disallow editing
        source_code.config(state=tk.DISABLED)

    def get_as_button(self, frame: tk.Frame, code_preview_frame: tk.Frame, execution_configuration) -> tk.Button:
        return tk.Button(
            frame,
            text=self.type + " @ " + self.values["start_line"],
            command=lambda: self.show_code_section(code_preview_frame, execution_configuration),
        )

    def __insert_pragmas(self, source_code: tk.Text, pragmas: List[Tuple[int, int, str]]):
        highlight_start_positions = []

        idx = 0
        for start_line, end_line, pragma_str in sorted(
            pragmas, reverse=True, key=lambda v: int(v[0])
        ):  # sort reverse by line num
            # add pragma string
            source_code.insert(str(start_line) + ".0", "    " + pragma_str + "\n")
            # highlight inserted pragmas and their target code sections
            pos = self.__highlight_code(
                source_code, start_line, end_line + 1, idx
            )  # +1 to account for added pragma line
            highlight_start_positions.append(pos)
            idx += 1

        return sorted(highlight_start_positions, key=lambda value: int(value.split(".")[0]))

    def __highlight_code(self, source_code: tk.Text, start_line: int, end_line: int, index):
        """highlights the specified lines in the source code preview and returns the used start position.
        index is used to determine the color."""
        end_line_length = 200
        background_color = "#e5f2b3" if index % 2 == 0 else "#a4ed9a"
        # highlight code
        start_pos = str(start_line) + ".0"
        end_pos = str(end_line) + "." + str(end_line_length)
        source_code.tag_add("start" + str(index), start_pos, end_pos)
        source_code.tag_config("start" + str(index), background=background_color, foreground="black")
        return start_pos

    def __get_pragmas(self) -> List[Tuple[int, int, str]]:
        """returns a list of source code lines and pragmas to be inserted into the code preview"""
        pragmas = []
        if self.type == "do_all" or self.type == "reduction":
            pragma = "#pragma omp parallel for "
            if len(self.values["first_private"]) > 0:
                pragma += "firstprivate(" + ",".join(self.values["first_private"]) + ") "
            if len(self.values["private"]) > 0:
                pragma += "private(" + ",".join(self.values["private"]) + ") "
            if len(self.values["last_private"]) > 0:
                pragma += "lastprivate(" + ",".join(self.values["last_private"]) + ") "
            if len(self.values["shared"]) > 0:
                pragma += "shared(" + ",".join(self.values["shared"]) + ") "
            if len(self.values["reduction"]) > 0:
                reductions_dict: Dict[Any, Any] = dict()
                for entry in self.values["reduction"]:
                    red_type = entry.split(":")[0]
                    var = entry.split(":")[1]
                    if red_type not in reductions_dict:
                        reductions_dict[red_type] = []
                    reductions_dict[red_type].append(var)
                for red_type in reductions_dict:
                    pragma += "reduction(" + red_type + ":" + ",".join(reductions_dict[red_type]) + ") "
            pragma_tuple = (self.start_line, self.end_line, pragma)
            pragmas.append(pragma_tuple)
            return pragmas
        elif self.type == "pipeline":
            for stage in self.values["stages"]:
                pragma = "#pragma omp task "
                if len(stage["first_private"]) > 0:
                    pragma += "firstprivate(" + ",".join(stage["first_private"]) + ") "
                if len(stage["private"]) > 0:
                    pragma += "private(" + ",".join(stage["private"]) + ") "
                if len(stage["shared"]) > 0:
                    pragma += "shared(" + ",".join(stage["shared"]) + ") "
                if len(stage["reduction"]) > 0:
                    reductions_dict = dict()
                    for entry in stage["reduction"]:
                        red_type = entry.split(":")[0]
                        var = entry.split(":")[1]
                        if red_type not in reductions_dict:
                            reductions_dict[red_type] = []
                        reductions_dict[red_type].append(var)
                    for red_type in reductions_dict:
                        pragma += "reduction(" + red_type + ":" + ",".join(reductions_dict[red_type]) + ") "
                if len(stage["in_deps"]) > 0:
                    pragma += "depends(in:" + ",".join(stage["in_deps"]) + ") "
                if len(stage["out_deps"]) > 0:
                    pragma += "depends(out:" + ",".join(stage["out_deps"]) + ") "
                if len(stage["in_out_deps"]) > 0:
                    pragma += "depends(inout:" + ",".join(stage["in_out_deps"]) + ") "
                pragma_tuple = (
                    int(stage["startsAtLine"].split(":")[1]),
                    int(stage["endsAtLine"].split(":")[1]),
                    pragma,
                )
                pragmas.append(pragma_tuple)
        else:
            pragmas.append(
                (
                    self.start_line,
                    self.end_line,
                    "#CURRENTLY UNSUPPORTED PREVIEW FOR TYPE: " + self.type,
                )
            )
        return pragmas

    def get_details(self) -> str:
        """Returns the details string which should be shown when hovering over the Suggestion button."""
        import pprint

        return pprint.pformat(self.values)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.


import json
from typing import Any, Dict, List
from discopop_explorer.PEGraphX import Node
from discopop_explorer.pattern_detectors.PatternBase import PatternBase

from discopop_library.PatternIdManagement.unique_pattern_id import get_unique_pattern_id
from discopop_library.discopop_optimizer.classes.context.Update import Update, construct_update_from_dict
from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum
from discopop_library.discopop_optimizer.classes.types.Aliases import DeviceID

SuggestionId = int


class OptimizerOutputPattern(PatternBase):
    applied_patterns: List[Dict[str, Any]]
    data_movement: List[Update]
    decisions: List[int]
    host_device_id: int

    def __init__(self, node: Node, decisions: List[int], host_device_id: int):
        PatternBase.__init__(self, node)
        self.applied_patterns = []
        self.data_movement = []
        self.decisions = decisions
        self.host_device_id = host_device_id

    def reconstruct_from_file(self, file_path: str):  # todo remove?
        with open(file_path, "r") as f:
            loaded_data = json.load(f)
        self.applied_patterns = loaded_data["applied_patterns"]
        self.pattern_id = loaded_data["pattern_id"]
        self.host_device_id = loaded_data["host_device_id"]

        for values in loaded_data["data_movement"]:
            self.data_movement.append(construct_update_from_dict(values))

    def dump_to_file(self, file_path: str):  # todo remove?
        dumpable_dict: Dict[str, Any] = dict()
        dumpable_dict["applied_patterns"] = self.applied_patterns
        dumpable_dict["data_movement"] = [update.toDict() for update in self.data_movement]
        dumpable_dict["pattern_id"] = self.pattern_id
        dumpable_dict["host_device_id"] = self.host_device_id

        with open(file_path, "w") as f:
            json.dump(dumpable_dict, f)

    def add_pattern(self, pattern_id: SuggestionId, target_device_id: DeviceID, device_type: DeviceTypeEnum):
        if self.pattern_id is None:
            # get a unique pattern_id
            self.pattern_id = get_unique_pattern_id()
        tmp_dict = {"pattern_id": pattern_id, "device_id": target_device_id, "device_type": device_type}
        self.applied_patterns.append(tmp_dict)

    def add_data_movement(self, update: Update):
        self.data_movement.append(update)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
from typing import List, Dict, Tuple, Optional, cast, Any

from discopop_explorer.PEGraphX import (
    CUNode,
    EdgeType,
    FunctionNode,
    Node,
    PEGraphX,
    NodeID,
    LineID,
)
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.task_parallelism.alias_detection import (
    get_alias_information as get_alias_detection_result,
)
from discopop_explorer.pattern_detectors.task_parallelism.classes import (
    TaskParallelismInfo,
    TPIType,
)
from discopop_explorer.pattern_detectors.task_parallelism.tp_utils import (
    line_contained_in_region,
    get_function_call_from_source_code,
    get_called_function_and_parameter_names_from_function_call,
    demangle,
    get_called_functions_recursively,
)


def detect_dependency_clauses_alias_based(
    pet: PEGraphX,
    suggestions: List[PatternInfo],
    file_mapping_path: str,
    cu_xml: str,
    dep_file: str,
    cu_inst_result_file: str,
    discopop_build_path: str,
) -> List[PatternInfo]:
    """Wrapper for alias based dependency detection.
    :param pet: PET Graph
    :param suggestions: List[PatternInfo]
    :param file_mapping_path: path to FileMapping file
    :param cu_xml: path to cu_xml file
    :param dep_file: path to dependency file
    :param cu_inst_result_file: path to CUInstResult.txt
    :param discopop_build_path: path to discopop build directory
    :return: List[PatternInfo]
    """
    # Read contents of file_mapping
    source_code_files = dict()
    with open(file_mapping_path) as f:
        for line in f.readlines():
            line = line.replace("\n", "")
            line_split = line.split("\t")
            source_code_files[line_split[0]] = line_split[1]
    # get RAW depencency information as a dict
    raw_dependency_information = get_raw_dependency_information_from_dep_file(dep_file)
    # parse cu_inst_result_file contents into dict
    cu_inst_result_dict = get_dict_from_cu_inst_result_file(cu_inst_result_file)
    aliases = get_alias_information(pet, suggestions, source_code_files)
    # get function-internal parameter aliases
    function_parameter_alias_dict = get_function_internal_parameter_aliases(
        file_mapping_path, cu_xml, discopop_build_path
    )
    # find dependencies between calls of different functions inside function scopes
    suggestions = identify_dependencies_for_different_functions(
        pet, suggestions, aliases, source_code_files, raw_dependency_information
    )
    # find dependencies between calls of same function inside function scopes
    suggestions = identify_dependencies_for_same_functions(
        pet, suggestions, source_code_files, cu_inst_result_dict, function_parameter_alias_dict
    )
    return suggestions


def get_raw_dependency_information_from_dep_file(dep_file: str) -> Dict[str, List[Tuple[str, str]]]:
    """return RAW dependency information contained in dep_file in the form of a dictionary.
    Format: {source_line: [(sink_line, var_name)]
    :param dep_file: path to dependency file
    :return: RAW dictionary
    """
    raw_dependencies: Dict[str, List[Tuple[str, str]]] = dict()
    with open(dep_file) as f:
        for line in f.readlines():
            line = line.replace("\n", "")
            # format of dependency entries in _dep.txt-file:
            #   sourceLine NOM RAW sinkLine|variable
            if " NOM " not in line:
                continue
            split_line = line.split(" NOM ")
            source_line = split_line[0]
            # split entries
            entries = []
            current_entry = ""
            for word in split_line[1].split(" "):
                word = word.replace(" ", "")
                if word == "RAW" or word == "WAR" or word == "WAW" or word == "INIT":
                    if len(current_entry) > 0:
                        entries.append(current_entry)
                    current_entry = ""
                if len(current_entry) > 0:
                    current_entry += " " + word
                else:
                    current_entry += word
            if len(current_entry) > 0:
                entries.append(current_entry)
            if source_line not in raw_dependencies:
                raw_dependencies[source_line] = []
            for entry in entries:
                # filter for RAW dependencies
                split_entry = entry.split(" ")
                if split_entry[0] != "RAW":
                    continue
                split_sink_line_var = split_entry[1].split("|")
                sink_line = split_sink_line_var[0]
                var_name = split_sink_line_var[1].replace(".addr", "")
                raw_dependencies[source_line].append((sink_line, var_name))
    return raw_dependencies


def get_dict_from_cu_inst_result_file(
    cu_inst_result_file: str,
) -> Dict[str, List[Dict[str, Optional[str]]]]:
    """Parses the information contained in cu_inst_result_file into a dictionary of dictionaries,
    ordered by dependency type and returns the dictionary.
    :param cu_inst_result_file: Path (string) to cu_inst_result file
    :return: parsed dictionary"""
    res_dict: Dict[str, List[Dict[str, Optional[str]]]] = {"RAW": [], "WAR": [], "WAW": []}
    try:
        with open(cu_inst_result_file) as f:
            for line in f:
                line = line.replace("\n", "")
                line_split = line.split(" ")
                dep_type = line_split[0]
                target_type = line_split[2]
                if target_type == "line:":
                    target_function = None
                    target_line = line_split[3]
                    target_var = line_split[5]
                elif target_type == "function:":
                    target_function = line_split[3]
                    target_line = line_split[5]
                    target_var = line_split[7]
                else:
                    raise ValueError("Unknown type: ", target_type)
                cur_line_dict = {
                    "function": target_function,
                    "line": target_line.replace(",", ""),
                    "var": target_var,
                }
                res_dict[dep_type].append(cur_line_dict)
    except FileNotFoundError as ex:
        raise ex
    return res_dict


def get_alias_information(pet: PEGraphX, suggestions: List[PatternInfo], source_code_files: Dict[str, str]):
    """Generate and return alias information dictionary.
    :param pet: PET Graph
    :param suggestions: List[PatternInfo]
    :param source_code_files: File-Mapping dictionary
    :return: alias information dictionary
    """
    # iterate over task suggestions
    task_suggestions = [s for s in [e for e in suggestions if type(e) == TaskParallelismInfo] if s.type is TPIType.TASK]
    # collect alias information
    aliases: Dict[TaskParallelismInfo, List[List[Tuple[str, str, LineID, LineID]]]] = dict()
    called_function_cache: Dict[Any, Any] = dict()
    for ts in task_suggestions:
        current_alias_entry = []
        potential_parent_functions = __get_potential_parent_functions(pet, ts)
        # get parent function
        for parent_function in potential_parent_functions:
            # get recursive function call from original source code
            try:
                function_call_string = get_function_call_from_source_code(
                    source_code_files, int(ts.pragma_line), ts.node_id.split(":")[0]
                )
            except IndexError:
                continue
            # get function parameter names from recursive function call
            (
                function_name,
                parameter_names,
            ) = get_called_function_and_parameter_names_from_function_call(
                function_call_string, ts._node.recursive_function_calls[0], cast(CUNode, ts._node)
            )
            # get CU Node object of called function
            called_function_cu_id: Optional[NodeID] = None
            for recursive_function_call_entry in ts._node.recursive_function_calls:
                if recursive_function_call_entry is None:
                    continue
                if "," in recursive_function_call_entry:
                    recursive_function_call_entry = recursive_function_call_entry.split(",")[0]
                recursive_function_call_entry_split = recursive_function_call_entry.split(" ")
                recursive_function_call_line = recursive_function_call_entry_split[1]
                if int(recursive_function_call_line.split(":")[1]) == int(ts.pragma_line):
                    # correct function call found
                    # find corresponding function CU
                    for tmp_func_cu in pet.all_nodes(FunctionNode):
                        if tmp_func_cu.name == recursive_function_call_entry_split[0]:
                            called_function_cu_id = tmp_func_cu.id
            # get aliases for parameters
            for idx, param in enumerate(parameter_names):
                if param is None:
                    continue
                if called_function_cu_id is None:
                    continue
                current_alias = [
                    (
                        param,
                        parent_function.name,
                        parent_function.start_position(),
                        parent_function.end_position(),
                    )
                ]
                current_alias += get_alias_for_parameter_at_position(
                    pet,
                    cast(FunctionNode, pet.node_at(called_function_cu_id)),
                    idx,
                    source_code_files,
                    [],
                    called_function_cache,
                )
                current_alias_entry.append(current_alias)
        aliases[ts] = current_alias_entry
    # join aliases on first element (original identifier)
    for key in aliases:
        joined_aliases = []
        while aliases[key]:
            join_on = aliases[key].pop()
            join_indices = []
            for idx, alias_entry in enumerate(aliases[key]):
                if alias_entry[0] == join_on[0]:
                    join_indices.append(idx)
            # sort reversed to prevent errors due to popping elements
            join_indices.sort(reverse=True)
            for idx in join_indices:
                to_be_joined = aliases[key].pop(idx)
                to_be_joined.pop(0)
                join_on += to_be_joined
            joined_aliases.append(join_on)
        aliases[key] = joined_aliases
    return aliases


def get_function_internal_parameter_aliases(
    file_mapping_path: str, cu_xml_path: str, discopop_build_path: str
) -> Dict[str, List[Tuple[str, str]]]:
    """Wrapper to execute simple alias analysis and parse results into dict (function name to list of alias-tuples).
    :param file_mapping_path: path to filemapping file
    :param cu_xml_path: path to cu_xml file
    :param discopop_build_path: path to discopop build directory
    :result: function-internal alias detection results in dict form"""
    # execute simple alias detection
    alias_detection_temp_file = os.getcwd() + "/alias_detection_temp.txt"
    # get absolute file paths
    file_mapping_path = os.path.abspath(file_mapping_path)
    cu_xml_path = os.path.abspath(cu_xml_path)

    # execute simple alias detection
    alias_detection_result = get_alias_detection_result(
        file_mapping_path, cu_xml_path, alias_detection_temp_file, discopop_build_path
    )

    # check if alias_detection_result has contents
    if len(alias_detection_result) == 0:
        return dict()
    # create dict:
    alias_dict: Dict[str, List[Tuple[str, str]]] = dict()
    for line in alias_detection_result.split("\n"):
        line = line.replace("\n", "")
        if ";" not in line:
            continue
        line_split = line.split(";")
        fname = line_split[1]
        var_name = line_split[2]
        alias_name = line_split[3]
        if var_name == alias_name:
            continue
        if fname not in alias_dict:
            alias_dict[fname] = []
        alias_dict[fname].append((var_name, alias_name))
    return alias_dict


def identify_dependencies_for_different_functions(
    pet: PEGraphX,
    suggestions: List[PatternInfo],
    aliases: Dict[TaskParallelismInfo, List[List[Tuple[str, str, LineID, LineID]]]],
    source_code_files: Dict[str, str],
    raw_dependency_information: Dict[str, List[Tuple[str, str]]],
) -> List[PatternInfo]:
    """Identify dependency clauses for all combinations of suggested tasks concerning different called functions
     and supplement the suggestions.
    :param pet: PET Graph
    :param suggestions: List[PatternInfo]
    :param aliases: alias information dict
    :param source_code_files: File-Mapping dictionary
    :param raw_dependency_information: RAW information dict
    :return: List[PatternInfo]
    """
    # wrapper to start __check_dependence_of_task_pair for all viable combinations of suggested tasks
    result_suggestions: List[PatternInfo] = []
    task_suggestions: List[TaskParallelismInfo] = []
    for s in suggestions:
        if type(s) == TaskParallelismInfo:
            if s.type is TPIType.TASK:
                task_suggestions.append(s)
            else:
                result_suggestions.append(s)
        else:
            result_suggestions.append(s)
    # iterate over all combinations of tasks, ts_1 has to come before ts_2
    # get in and out dependencies to insert
    out_dep_updates: Dict[TaskParallelismInfo, List[str]] = dict()
    in_dep_updates: Dict[TaskParallelismInfo, List[str]] = dict()
    for ts_1 in task_suggestions:
        # get parent function
        potential_parent_functions_1 = __get_potential_parent_functions(pet, ts_1)
        while potential_parent_functions_1:
            potential_parent_functions_1.pop()
            # get recursive function call from original source code
            try:
                function_call_string_1 = get_function_call_from_source_code(
                    source_code_files, int(ts_1.pragma_line), ts_1.node_id.split(":")[0]
                )
            except IndexError:
                continue
            # get function parameter names from recursive function call
            (
                function_name_1,
                parameter_names_1,
            ) = get_called_function_and_parameter_names_from_function_call(
                function_call_string_1,
                ts_1._node.recursive_function_calls[0],
                cast(CUNode, ts_1._node),
            )
            for ts_2 in [s for s in task_suggestions if not s == ts_1]:
                # get parent function
                potential_parent_functions_2 = __get_potential_parent_functions(pet, ts_2)
                while potential_parent_functions_2:
                    potential_parent_functions_2.pop()
                    # get recursive function call from original source code
                    try:
                        function_call_string_2 = get_function_call_from_source_code(
                            source_code_files, int(ts_2.pragma_line), ts_2.node_id.split(":")[0]
                        )
                        (
                            function_name_2,
                            parameter_names_2,
                        ) = get_called_function_and_parameter_names_from_function_call(
                            function_call_string_2,
                            ts_2._node.recursive_function_calls[0],
                            cast(CUNode, ts_2._node),
                        )
                        # exclude pairs of same function from dependency detection
                        if function_name_1 == function_name_2:
                            continue
                    except IndexError:
                        continue
                    # get function parameter names from recursive function call
                    dependencies = check_dependence_of_task_pair(
                        aliases, raw_dependency_information, ts_1, parameter_names_1, ts_2
                    )

                    # check if task suggestion_1 occurs prior to task_suggestion_2
                    if ts_1._node.start_position().split(":")[0] == ts_2._node.start_position().split(":")[0]:
                        # same file id
                        ts_1_pragma_line = (
                            int(ts_1.pragma_line)
                            if ":" not in ts_1.pragma_line
                            else int(ts_1.pragma_line.split(":")[1])
                        )
                        ts_2_pragma_line = (
                            int(ts_2.pragma_line)
                            if ":" not in ts_2.pragma_line
                            else int(ts_2.pragma_line.split(":")[1])
                        )
                        # check line numbers
                        if ts_2_pragma_line < ts_1_pragma_line:
                            continue

                    for dependence_var in dependencies:
                        # Mark the variable as depend out for the first function and depend in for the second function.
                        if ts_1 not in out_dep_updates:
                            out_dep_updates[ts_1] = []
                        out_dep_updates[ts_1].append(dependence_var)
                        if ts_2 not in in_dep_updates:
                            in_dep_updates[ts_2] = []
                        in_dep_updates[ts_2].append(dependence_var)
    # perform updates of in and out dependencies
    for ts in task_suggestions:
        if ts in out_dep_updates:
            for out_dep_var in out_dep_updates[ts]:
                ts.out_dep.append(out_dep_var)
            ts.out_dep = list(set(ts.out_dep))
        if ts in in_dep_updates:
            for in_dep_var in in_dep_updates[ts]:
                ts.in_dep.append(in_dep_var)
            ts.in_dep = list(set(ts.in_dep))
        result_suggestions.append(ts)

    return result_suggestions


def __get_potential_parent_functions(pet: PEGraphX, sug: TaskParallelismInfo) -> List[Node]:
    """Helper function for identify_dependencies_for_same_functions.
    Creates a list of potential parents (Function CU Nodes) for a given suggestion.
    :param pet: PET Graph
    :param sug: target suggestion
    :return: List of potential parents of sug (Function CU Nodes)"""
    potential_parent_functions = [
        pet.node_at(e[0])
        for e in pet.in_edges(sug._node.id, EdgeType.CHILD)
        if isinstance(pet.node_at(e[0]), FunctionNode)
    ]
    if not potential_parent_functions:
        # perform BFS search on incoming CHILD edges to find closest parent function,
        # i.e. function which contains the CU.
        queue = [pet.node_at(e[0]) for e in pet.in_edges(sug._node.id, EdgeType.CHILD)]
        found_parent = None
        while len(queue) > 0 or not found_parent:
            current = queue.pop(0)
            if isinstance(current, FunctionNode):
                found_parent = current
                break
            queue += [pet.node_at(e[0]) for e in pet.in_edges(current.id, EdgeType.CHILD)]
        potential_parent_functions = [found_parent]
    return potential_parent_functions


def __get_potential_children_of_function(pet: PEGraphX, parent_function: Node) -> List[Node]:
    """Helper function for identify_dependencies_for_same_functions.
    Creates a list of CUNodes corresponding to the body of the given function.
    :param pet: PET Graph
    :param parent_function: function to be analyzed
    :return: List of CUNodes contained in functions body."""
    queue = pet.direct_children_or_called_nodes(parent_function)
    potential_children = []
    visited = []
    while queue:
        cur_potential_child = queue.pop()
        visited.append(cur_potential_child)
        # test if cur_potential_child is inside cur_potential_parent_functions scope
        if line_contained_in_region(
            cur_potential_child.start_position(),
            parent_function.start_position(),
            parent_function.end_position(),
        ) and line_contained_in_region(
            cur_potential_child.end_position(),
            parent_function.start_position(),
            parent_function.end_position(),
        ):
            potential_children.append(cur_potential_child)
        for tmp_child in pet.direct_children_or_called_nodes(cur_potential_child):
            if tmp_child not in queue and tmp_child not in potential_children and tmp_child not in visited:
                queue.append(tmp_child)
    return potential_children


def __get_recursive_calls_from_function(potential_children: List[Node], parent_function: Node) -> List[str]:
    """Helper function for identify_dependencies_for_same_functions.
    Creates a list of recursive function calls located inside the body of the given function.
    :param potential_children: List of CUs contained in function's body
    :param parent_function: Parent function
    :return: List of recursive function call contained in parent_function's body
    """
    recursive_function_calls = []
    for child in [
        c
        for c in potential_children
        if line_contained_in_region(
            c.start_position(), parent_function.start_position(), parent_function.end_position()
        )
        and line_contained_in_region(c.end_position(), parent_function.start_position(), parent_function.end_position())
    ]:
        for e in child.recursive_function_calls:
            if e is not None:
                recursive_function_calls.append(e)
    recursive_function_calls = list(set(recursive_function_calls))
    return recursive_function_calls


def identify_dependencies_for_same_functions(
    pet: PEGraphX,
    suggestions: List[PatternInfo],
    source_code_files: Dict[str, str],
    cu_inst_result_dict: Dict[str, List[Dict[str, Optional[str]]]],
    function_parameter_alias_dict: Dict[str, List[Tuple[str, str]]],
) -> List[PatternInfo]:
    """Identify dependency clauses for all combinations of suggested tasks concerning equal called functions
    and supplement the suggestions.
    :param pet: PET Graph
    :param suggestions: List[PatternInfo]
    :param source_code_files: File-Mapping dictionary
    :param cu_inst_result_dict: CUInstResult.txt information dict
    :param function_parameter_alias_dict: results of function-internal alias detection in dict form
    :return: List[PatternInfo]"""
    # Idea:
    # 1. iterate over task suggestions
    # 2. get parent function (pf) and called function (cf)
    # 3. get R/W information for cf's parameters based on CUInstResult.txt
    # 4. iterate over successive calls to cf, named scf
    # 5. get R/W information for scf's parameters based on CUInstResult.txt
    # 6. check cf's R/W information against scf's R/W information and identify dependencies
    # 6.1 Intersect cf's parameters with scf's parameters
    # 6.2 get task suggestion corresponding to scf
    # 6.3 If intersecting parameter of cf is RAW, add dependency (scf:in, cf:out)

    result_suggestions: List[PatternInfo] = []
    task_suggestions: List[TaskParallelismInfo] = []
    for s in suggestions:
        if type(s) == TaskParallelismInfo:
            if s.type is TPIType.TASK:
                task_suggestions.append(s)
            else:
                result_suggestions.append(s)
        else:
            result_suggestions.append(s)

    out_dep_updates: Dict[TaskParallelismInfo, List[Tuple[str, bool]]] = dict()
    in_dep_updates: Dict[TaskParallelismInfo, List[Tuple[str, bool]]] = dict()
    function_raw_information_cache: Dict[str, List[Tuple[bool, bool]]] = dict()
    # 1. iterate over task suggestions
    for ts_1 in task_suggestions:
        # 2. get parent function (pf) and called function (cf)
        # get parent function
        potential_parent_functions_1 = __get_potential_parent_functions(pet, ts_1)
        while potential_parent_functions_1:
            cur_potential_parent_function = potential_parent_functions_1.pop()
            # get recursive function call from original source code
            try:
                get_function_call_from_source_code(source_code_files, int(ts_1.pragma_line), ts_1.node_id.split(":")[0])
            except IndexError:
                continue
            # get potential children by dfs enumerating children nodes inside cur_potential_parent_functions scope
            potential_children = __get_potential_children_of_function(pet, cur_potential_parent_function)

            cppf_recursive_function_calls = __get_recursive_calls_from_function(
                potential_children, cur_potential_parent_function
            )
            outer_breaker = False
            for rfce_idx_1, recursive_function_call_entry_1 in enumerate(cppf_recursive_function_calls):
                if outer_breaker:
                    break
                # 3. get R/W information for cf's parameters based on CUInstResult.txt
                called_function_name_1, call_line_1 = recursive_function_call_entry_1.split(",")[0].split(" ")
                lower_line_num_1 = ts_1.pragma_line
                if ":" in lower_line_num_1:
                    lower_line_num_1 = lower_line_num_1.split(":")[1]
                lower_line_num_1 = int(lower_line_num_1)
                ret_val_1 = get_function_call_parameter_rw_information(
                    pet,
                    call_line_1,
                    ts_1._node,
                    lower_line_num_1,
                    True,
                    False,
                    cu_inst_result_dict,
                    source_code_files,
                    [],
                    function_raw_information_cache,
                    function_parameter_alias_dict,
                    called_function_name=called_function_name_1,
                )
                if ret_val_1 is None:
                    continue
                (
                    recursive_function_call_line_1,
                    parameter_names_1_raw_information,
                    recursively_visited_1,
                    function_raw_information_cache,
                ) = ret_val_1
                # 4. iterate over successive calls to cf, named scf
                for rfce_idx_2, recursive_function_call_entry_2 in enumerate(cppf_recursive_function_calls):
                    if rfce_idx_2 == rfce_idx_1:
                        continue
                    if recursive_function_call_entry_2 is None:
                        continue
                    # 5. get R/W Information for scf
                    called_function_name_2, call_line_2 = recursive_function_call_entry_2.split(",")[0].split(" ")
                    lower_line_num_2 = ts_1.pragma_line
                    if ":" in lower_line_num_2:
                        lower_line_num_2 = lower_line_num_2.split(":")[1]
                    lower_line_num_2 = int(lower_line_num_2)
                    ret_val_2 = get_function_call_parameter_rw_information(
                        pet,
                        call_line_2,
                        ts_1._node,
                        lower_line_num_2,
                        False,
                        True,
                        cu_inst_result_dict,
                        source_code_files,
                        [],
                        function_raw_information_cache,
                        function_parameter_alias_dict,
                        called_function_name=called_function_name_2,
                    )
                    if ret_val_2 is None:
                        continue
                    (
                        recursive_function_call_line_2,
                        parameter_names_2_raw_information,
                        recursively_visited_2,
                        function_raw_information_cache,
                    ) = ret_val_2

                    # 6. check cf's R/W information against scf's R/W information and identify dependencies
                    # 6.1 Intersect cf's parameters with scf's parameters
                    intersection = []
                    for param_entry_1 in parameter_names_1_raw_information:
                        if param_entry_1[0] is None:
                            continue
                        for param_entry_2 in parameter_names_2_raw_information:
                            if param_entry_1[0] == param_entry_2[0]:
                                # filter out potential numbers as variable names
                                try:
                                    int(param_entry_1[0])
                                except ValueError:
                                    intersection.append(param_entry_1)
                    intersection = list(set(intersection))
                    # 6.2 get task suggestion corresponding to scf
                    for ts_2 in task_suggestions:
                        if ts_2 == ts_1:
                            continue
                        if ts_2.pragma_line != recursive_function_call_line_2.split(":")[1]:
                            continue

                        # 6.3 If intersecting parameter of cf is RAW, add dependency (scf:in, cf:out)
                        for intersection_var, is_pessimistic in [(e[0], e[2]) for e in intersection if e[1]]:
                            if ts_1 not in out_dep_updates:
                                out_dep_updates[ts_1] = []
                            out_dep_updates[ts_1].append((intersection_var, is_pessimistic))
                            if ts_2 not in in_dep_updates:
                                in_dep_updates[ts_2] = []
                            in_dep_updates[ts_2].append((intersection_var, is_pessimistic))
                    outer_breaker = True
    # perform updates of in and out dependencies
    return __perform_dependency_updates(task_suggestions, in_dep_updates, out_dep_updates, result_suggestions)


def __perform_dependency_updates(
    task_suggestions: List[TaskParallelismInfo],
    in_dep_updates: Dict[TaskParallelismInfo, List[Tuple[str, bool]]],
    out_dep_updates: Dict[TaskParallelismInfo, List[Tuple[str, bool]]],
    result_suggestions: List[PatternInfo],
) -> List[PatternInfo]:
    """Helper function for identify_dependencies_for_same_functions.
    Applies the identified updates to the in and out dependency lists of each task suggestion.
    Returns the list of modified suggestions.
    :param task_suggestions: list of task suggestions
    :param in_dep_updates: updates to be done to in-dependencies
    :param out_dep_updates: updates to be done to out-dependencies
    :param result_suggestions: resulting suggestions as created by identify_dependencies_for_same_functions
        prior to calling this function
    :return: modified result_suggestions list"""
    for ts in task_suggestions:
        if ts in out_dep_updates:
            for out_dep_var, is_pessimistic in out_dep_updates[ts]:
                if out_dep_var not in ts.out_dep and is_pessimistic:
                    print(
                        "TPDet: Warning: Pessimistic Dependency:: CUid:",
                        ts.node_id,
                        " Type: OUT  VarName:",
                        out_dep_var,
                    )
                ts.out_dep.append(out_dep_var)
            ts.out_dep = list(set(ts.out_dep))
        if ts in in_dep_updates:
            for in_dep_var, is_pessimistic in in_dep_updates[ts]:
                if in_dep_var not in ts.in_dep and is_pessimistic:
                    print(
                        "TPDet: Warning: Pessimistic Dependency:: CUid:",
                        ts.node_id,
                        " Type: IN  VarName:",
                        in_dep_var,
                    )
                ts.in_dep.append(in_dep_var)
            ts.in_dep = list(set(ts.in_dep))
        result_suggestions.append(ts)
    return result_suggestions


def get_alias_for_parameter_at_position(
    pet: PEGraphX,
    function: FunctionNode,
    parameter_position: int,
    source_code_files: Dict[str, str],
    visited: List[Tuple[Node, int]],
    called_function_cache: Dict[Any, Any],
) -> List[Tuple[str, str, LineID, LineID]]:
    """Returns alias information for a parameter at a specific position.
    :param pet: PET Graph
    :param function: CUNode of called function
    :param parameter_position: position of the parameter to be analyzed
    :param source_code_files: File-Mapping dictionary
    :param visited: List of already traversed function-index-combinations
    :param called_function_cache: cache for results of tp_utils.get_called_functions_recursively()
    :return: alias information for the specified parameter
    """
    visited.append((function, parameter_position))
    parameter_name = function.args[parameter_position].name
    # append Alias information for parameter to result
    result = [(parameter_name, function.name, function.start_position(), function.end_position())]

    # find function calls which use the parameter
    # iterate over CUs
    for cu in [pet.node_at(cuid) for cuid in [e[1] for e in pet.out_edges(function.id)]]:
        # iterate over children of CU and retrieve called functions
        called_functions = get_called_functions_recursively(pet, cu, [], called_function_cache)
        called_functions = list(set(called_functions))
        # iterate over called functions
        for called_function in called_functions:
            # read line from source code (iterate over lines of CU to search for function call)
            for line in range(cu.start_line, cu.end_line + 1):
                try:
                    source_code_line = get_function_call_from_source_code(source_code_files, line, cu.id.split(":")[0])
                except IndexError:
                    continue
                # get parameter names from call

                (
                    function_name,
                    call_parameters,
                ) = get_called_function_and_parameter_names_from_function_call(
                    source_code_line, called_function.name, cast(CUNode, cu)
                )
                # check if parameter_name is contained
                for idx, pn in enumerate(call_parameters):
                    if pn == parameter_name:
                        # check if same configuration for alias detection has been used:
                        if (called_function, idx) not in visited:
                            # if not, start recursion
                            result += get_alias_for_parameter_at_position(
                                pet,
                                cast(FunctionNode, called_function),
                                idx,
                                source_code_files,
                                visited,
                                called_function_cache,
                            )
    return result


def check_dependence_of_task_pair(
    aliases: Dict[TaskParallelismInfo, List[List[Tuple[str, str, LineID, LineID]]]],
    raw_dependency_information: Dict[str, List[Tuple[str, str]]],
    task_suggestion_1: TaskParallelismInfo,
    param_names_1: List[Optional[str]],
    task_suggestion_2: TaskParallelismInfo,
) -> List[str]:
    """Check if function calls specified by task_suggestion_1 and _2 are dependent and
    return a list of found dependencies. An empty list is returned if task_suggestion_2 occurs before task_suggestion_1.
    :param aliases: alias information dict
    :param raw_dependency_information: RAW information dict
    :param task_suggestion_1: first suggestion for the check
    :param param_names_1: List of parameter names used in the function call specified by task_suggestion_1
    :param task_suggestion_2: second suggestion for the check
    :return: List of found dependencies
    """
    dependencies = []
    # iterate over parameters of task_1
    for parameter in param_names_1:
        if parameter is None:
            continue
        # get aliases for parameter
        for alias_entry in aliases[task_suggestion_1]:
            # skip wrong alias entries
            if not alias_entry[0][0] == parameter:
                continue
            # intersect alias_entry of task_suggestion_1 with entries of task_suggestion_2
            alias_entries_2 = []
            for alias_entry_2 in aliases[task_suggestion_2]:
                alias_entries_2 += alias_entry_2
            intersection = list(set([ae for ae in alias_entry if ae in alias_entries_2]))
            # get sink lines
            # (start and end line of task_sug_1's parent func)
            sink_lines_start = alias_entry[0][2].split(":")
            sink_lines_end = alias_entry[0][3].split(":")
            sink_lines = []
            for ln in range(int(sink_lines_start[1]), int(sink_lines_end[1]) + 1):
                sink_lines.append("" + sink_lines_start[0] + ":" + str(ln))
            # check if there is a RAW on the variable within sink lines and source lines.
            for intersection_entry in intersection:
                # get intersection variable and source lines
                intersecting_variable = intersection_entry[0]
                source_lines_start = intersection_entry[2].split(":")
                source_lines_end = intersection_entry[3].split(":")
                source_lines = []
                for ln in range(int(source_lines_start[1]), int(source_lines_end[1]) + 1):
                    source_lines.append("" + source_lines_start[0] + ":" + str(ln))
                # check if there is a RAW on the variable within sink lines and source lines.
                if source_lines == sink_lines:
                    continue
                for source_line in source_lines:
                    if source_line not in raw_dependency_information:
                        continue
                    for raw_dep_entry in raw_dependency_information[source_line]:
                        if raw_dep_entry[1] == intersecting_variable:
                            if raw_dep_entry[0] in sink_lines:
                                dependencies.append(parameter)
    dependencies = list(set(dependencies))
    # check if task suggestion_1 occurs prior to task_suggestion_2
    if task_suggestion_1._node.start_position().split(":")[0] == task_suggestion_2._node.start_position().split(":")[0]:
        # same file id
        ts_1_pragma_line = (
            int(task_suggestion_1.pragma_line)
            if ":" not in task_suggestion_1.pragma_line
            else int(task_suggestion_1.pragma_line.split(":")[1])
        )
        ts_2_pragma_line = (
            int(task_suggestion_2.pragma_line)
            if ":" not in task_suggestion_2.pragma_line
            else int(task_suggestion_2.pragma_line.split(":")[1])
        )
        # check line numbers
        if ts_2_pragma_line < ts_1_pragma_line:
            return []
    return dependencies


def get_function_call_parameter_rw_information(
    pet: PEGraphX,
    call_position: str,
    parent_cu_node: Node,
    lower_line_num: int,
    equal_lower_line_num: bool,
    greater_lower_line_num: bool,
    cu_inst_result_dict: Dict[str, List[Dict[str, Optional[str]]]],
    source_code_files: Dict[str, str],
    recursively_visited: List[Node],
    function_raw_information_cache: Dict[str, List[Tuple[bool, bool]]],
    function_parameter_alias_dict: Dict[str, List[Tuple[str, str]]],
    called_cu_id: Optional[str] = None,
    called_function_name: Optional[str] = None,
) -> Optional[Tuple[str, List[Tuple[str, bool, bool]], List[Node], Dict[str, List[Tuple[bool, bool]]]]]:
    """Retrieves the call_position and information whether the parameters of the target function are modified
    within the respective function, based on the contents of cu_inst_result_dict.
    Either called_cu_id, called_function_name or both need to be set.
    :param pet: PET Graph
    :param call_position: position of function call in source code (e.g. 14:275)
    :param parent_cu_node: CUNode corresponding to the CU Node containing the target function
    :param lower_line_num: called_function_cu_id detections is restricted to
        source code lines >= / > / == lower_line_num to ignore prior function calls.
        Different behavior is achieved by combining equal_lower_line and greater_lower_line
    :param equal_lower_line_num: allow called_function_cu_id detection for lines == lower_line_num
    :param greater_lower_line_num: allow called_function_cu_id detection for lines >= lower_line_num
    :param cu_inst_result_dict: Contents of the CUInst_Result.txt, converted into a dict.
    :param source_code_files: File Mapping dictionary
    :param recursively_visited: List of already visited CU Nodes to prevent endless recursions
    :param function_raw_information_cache: Cache containing a mapping of function names to a list of booleans,
        representing the parameters of a given function and the information
        whether a specific parameter is modified by the respective function.
    :param function_parameter_alias_dict: results of function-internal alias detection in dict form
    :param called_cu_id: ID of the called function´s CU node
    :param called_function_name: Name of the called function
    :return: None, if anything unpredicted happens.
             Otherwise, (call_position, parameter_names_raw_information,
             recursively_visited, function_raw_information_cache),
             with parameter_names_raw_infotation being a list of Tuples containing the parameter names of the
             called function and the information wheter the respective parameter is modified by the function.
    """
    ###############
    # 5. get R/W information for called function´s (cf) parameters based on CUInstResult.txt
    # 5.1.get CU object corresponding to cf
    # 5.2. get R/W information for cf's parameters based on CUInstResult.txt
    # 5.3 get function call corresponding to scf from source code
    # 5.4 start recursion step
    # 5.5 match variable names with gathered R/W information
    ################

    if called_cu_id is None and called_function_name is None:
        raise ValueError("Unsufficient information!")
    # 5. get R/W information for scf's parameters based on CUInstResult.txt
    # 5.1. get CU object corresponding to called function
    called_function_cu_id: Optional[NodeID] = None
    if called_cu_id is not None:
        called_function_cu_id = cast(NodeID, called_cu_id)
        if called_function_name is None:
            called_function_name = pet.node_at(called_function_cu_id).name
    else:
        # get cu id of called function
        if equal_lower_line_num and not greater_lower_line_num:
            if int(call_position.split(":")[1]) == lower_line_num:
                # correct function call found
                # find corresponding function CU
                for tmp_func_cu in pet.all_nodes(FunctionNode):
                    if tmp_func_cu.name == called_function_name:
                        called_function_cu_id = tmp_func_cu.id
        elif not equal_lower_line_num and greater_lower_line_num:
            if int(call_position.split(":")[1]) > lower_line_num:
                # correct function call found
                # find corresponding function CU
                for tmp_func_cu in pet.all_nodes(FunctionNode):
                    if tmp_func_cu.name == called_function_name:
                        called_function_cu_id = tmp_func_cu.id
        else:
            if int(call_position.split(":")[1]) >= lower_line_num:
                # correct function call found
                # find corresponding function CU
                for tmp_func_cu in pet.all_nodes(FunctionNode):
                    if tmp_func_cu.name == called_function_name:
                        called_function_cu_id = tmp_func_cu.id
        if called_function_cu_id is None:
            return None
    if called_function_name is None:
        raise ValueError("No valid called function could be found!")
    # 5.2. get R/W information for successive called function's parameters based on CUInstResult.txt
    called_function_cu = cast(FunctionNode, pet.node_at(called_function_cu_id))
    # get raw info concerning the scope of called_function_cu
    raw_info = cu_inst_result_dict["RAW"]
    filtered_raw_info = [
        e
        for e in raw_info
        if line_contained_in_region(
            cast(str, e["line"]),
            called_function_cu.start_position(),
            called_function_cu.end_position(),
        )
    ]
    # iterate over args positions and check if RAW is reported
    raw_reported_for_param_positions: List[Tuple[bool, bool]] = []
    for arg_var in called_function_cu.args:
        raw_reported = False
        for raw_entry in filtered_raw_info:
            if raw_entry["var"] is None:
                continue
            raw_entry_var = raw_entry["var"]
            if raw_entry_var.replace(".addr", "") == arg_var.name.replace(".addr", ""):
                raw_reported = True
        raw_reported_for_param_positions.append((raw_reported, False))
    # store results in cache
    if called_function_cu.name not in function_raw_information_cache:
        function_raw_information_cache[called_function_cu.name] = raw_reported_for_param_positions
    else:
        # perform update
        new_cache_line = []
        for idx in range(0, len(function_raw_information_cache[called_function_cu.name])):
            old_cache_line = function_raw_information_cache[called_function_cu.name]
            new_cache_line.append(old_cache_line[idx] or raw_reported_for_param_positions[idx])
        function_raw_information_cache[called_function_cu.name] = new_cache_line

    # 5.3 get function call corresponding to scf from source code
    try:
        function_call_string = get_function_call_from_source_code(
            source_code_files,
            int(call_position.split(":")[1]),
            call_position.split(":")[0],
            called_function_name=demangle(called_function_name).split("(")[0],
        )
    except IndexError:
        return None
    # get function parameter names from recursive function call
    function_name, parameter_names = get_called_function_and_parameter_names_from_function_call(
        function_call_string, called_function_name, cast(CUNode, parent_cu_node)
    )

    # 5.4. start recursion step
    res_called_function_raw_information: List[Tuple[bool, bool]] = []
    if called_function_cu not in recursively_visited:
        (
            recursively_visited,
            res_called_function_name,
            res_called_function_raw_information,
            function_raw_information_cache,
        ) = get_function_call_parameter_rw_information_recursion_step(
            pet,
            called_function_cu,
            recursively_visited,
            function_raw_information_cache,
            cu_inst_result_dict,
            function_parameter_alias_dict,
            source_code_files,
        )
        function_raw_information_cache[called_function_cu.name] = res_called_function_raw_information
    else:
        # read cache
        if called_function_cu.name in function_raw_information_cache:
            res_called_function_raw_information = function_raw_information_cache[called_function_cu.name]

    # 5.5 match parameter_names with gathered R/W information of argument positions
    if len(raw_reported_for_param_positions) != len(parameter_names):
        return None
    parameter_names_raw_information: List[Tuple[str, bool, bool]] = []
    if len(raw_reported_for_param_positions) == len(res_called_function_raw_information):
        for idx in range(0, len(parameter_names)):
            tmp = (
                parameter_names[idx],
                (raw_reported_for_param_positions[idx][0] or res_called_function_raw_information[idx][0]),
                res_called_function_raw_information[idx][1],
            )
            if tmp[0] is None:
                continue
            tmp_not_none = cast(Tuple[str, bool, bool], tmp)
            parameter_names_raw_information.append(tmp_not_none)
        # overwrite cache
        new_cache_line = []
        for idx in range(0, len(function_raw_information_cache[called_function_cu.name])):
            old_cache_line = function_raw_information_cache[called_function_cu.name]
            update = [elem[1] for elem in parameter_names_raw_information]
            new_cache_line.append(old_cache_line[idx] or update[idx])
        function_raw_information_cache[called_function_cu.name] = new_cache_line
    else:
        # ignore recursion results
        for idx in range(0, len(parameter_names)):
            tmp = (parameter_names[idx], raw_reported_for_param_positions[idx][0], False)
            if tmp[0] is None:
                continue
            tmp_not_none = cast(Tuple[str, bool, bool], tmp)
            parameter_names_raw_information.append(tmp_not_none)

    return (
        call_position,
        parameter_names_raw_information,
        recursively_visited,
        function_raw_information_cache,
    )


def get_function_call_parameter_rw_information_recursion_step(
    pet: PEGraphX,
    called_function_cu: FunctionNode,
    recursively_visited: List[Node],
    function_raw_information_cache: Dict[str, List[Tuple[bool, bool]]],
    cu_inst_result_dict: Dict[str, List[Dict[str, Optional[str]]]],
    function_parameter_alias_dict: Dict[str, List[Tuple[str, str]]],
    source_code_files: Dict[str, str],
) -> Tuple[List[Node], str, List[Tuple[bool, bool]], Dict[str, List[Tuple[bool, bool]]]]:
    """Wrapper to execute __get_function_call_parameter_rw_information recursively,
    i.e. for every function call in called functions body.
    The gathered information is aggregated via a logical disjunction on a per-variable level.
    Results only contain the information whether a given variable is modified at some point,
    the specific location is not included.
    :param pet: PET Graph
    :param called_function_cu: CU Node corresponding to the called function to be checked
    :param recursively_visited: List of already visited CU Nodes to prevent endless recursions
    :param function_raw_information_cache: Cache containing a mapping of function names to a list of booleans,
        representing the parameters of a given function and the information
        whether a specific parameter is modified by the respective function.
    :param cu_inst_result_dict: Contents of the CUInst_Result.txt, converted into a dict.
    :param function_parameter_alias_dict: results of function-internal alias detection in dict form
    :param source_code_files: File-Mapping dictionary
    :return: (recursively_visited, called_function_name,
              called_function_args_raw_information, function_raw_information_cache)
    """

    # get potential children of called function
    recursively_visited.append(called_function_cu)
    queue_1 = pet.direct_children_or_called_nodes(called_function_cu)
    potential_children = []
    visited = []
    while queue_1:
        cur_potential_child = queue_1.pop()
        visited.append(cur_potential_child)
        # test if cur_potential_child is inside cur_potential_parent_functions scope
        if line_contained_in_region(
            cur_potential_child.start_position(),
            called_function_cu.start_position(),
            called_function_cu.end_position(),
        ) and line_contained_in_region(
            cur_potential_child.end_position(),
            called_function_cu.start_position(),
            called_function_cu.end_position(),
        ):
            if cur_potential_child not in potential_children:
                potential_children.append(cur_potential_child)
        for tmp_child in pet.direct_children_or_called_nodes(cur_potential_child):
            if tmp_child not in queue_1 and tmp_child not in potential_children and tmp_child not in visited:
                queue_1.append(tmp_child)

    called_function_args_raw_information = []
    for var in called_function_cu.args:
        called_function_args_raw_information.append((var.name, False, False))

    for child in [
        c
        for c in potential_children
        if line_contained_in_region(
            c.start_position(),
            called_function_cu.start_position(),
            called_function_cu.end_position(),
        )
        and line_contained_in_region(
            c.end_position(), called_function_cu.start_position(), called_function_cu.end_position()
        )
    ]:
        # find called functions
        for child_func in pet.direct_children_or_called_nodes_of_type(child, FunctionNode):
            # apply __get_function_call_parameter_rw_information
            if child not in recursively_visited:
                ret_val = get_function_call_parameter_rw_information(
                    pet,
                    child.start_position(),
                    child,
                    int(child.start_position().split(":")[1]),
                    True,
                    True,
                    cu_inst_result_dict,
                    source_code_files,
                    recursively_visited,
                    function_raw_information_cache,
                    function_parameter_alias_dict,
                    called_function_name=child_func.name,
                )
                if ret_val is None:
                    continue
                (
                    recursive_function_call_line,
                    parameter_names_raw_information,
                    recursively_visited,
                    function_raw_information_cache,
                ) = ret_val
                # perform or-conjunction of RAW information parent <> child
                for (
                    child_var_name,
                    child_raw_info,
                    child_is_pessimistic,
                ) in parameter_names_raw_information:
                    for idx, (var_name, raw_info, is_pessimistic) in enumerate(called_function_args_raw_information):
                        if var_name == child_var_name:
                            called_function_args_raw_information[idx] = (
                                var_name,
                                raw_info or child_raw_info,
                                child_is_pessimistic or is_pessimistic,
                            )

    # if parameter alias entry for parent function exists:
    if called_function_cu.name in function_parameter_alias_dict:
        alias_entries = function_parameter_alias_dict[called_function_cu.name]
        for var_name, alias_name in alias_entries:
            var_name_is_modified = False
            # check if alias_name occurs in any depencendy in any of called_function_cu's children,
            # recursively visits all children cu nodes in function body.
            function_internal_cu_nodes: List[Node] = []
            queue: List[Node] = [called_function_cu]
            while len(queue) > 0:
                cur: Node = queue.pop(0)
                # check if cur inside function body, append to function_internal_cu_nodes if so
                if line_contained_in_region(
                    cur.start_position(),
                    called_function_cu.start_position(),
                    called_function_cu.end_position(),
                ) and line_contained_in_region(
                    cur.end_position(),
                    called_function_cu.start_position(),
                    called_function_cu.end_position(),
                ):
                    function_internal_cu_nodes.append(cur)
                # add children to queue
                for cur_child in pet.direct_children_or_called_nodes(cur):
                    if cur_child not in function_internal_cu_nodes and cur_child not in queue:
                        queue.append(cur_child)
            for child_cu in function_internal_cu_nodes:
                child_in_deps = pet.in_edges(child_cu.id, EdgeType.DATA)
                child_out_deps = pet.out_edges(child_cu.id, EdgeType.DATA)
                dep_var_names = [x[2].var_name for x in child_in_deps + child_out_deps]
                dep_var_names_not_none = [x for x in dep_var_names if x is not None]
                dep_var_names_not_none = [x.replace(".addr", "") for x in dep_var_names_not_none]
                if alias_name in dep_var_names_not_none:
                    var_name_is_modified = True
                    break
            if var_name_is_modified:
                # update RAW information
                for idx, (old_var_name, raw_info, _) in enumerate(called_function_args_raw_information):
                    if old_var_name == var_name:
                        if not raw_info:
                            called_function_args_raw_information[idx] = (old_var_name, True, True)
                            # second True denotes the pessimistic nature of a potential created dependency
    # remove names from called_function_args_raw_information
    called_function_args_raw_information_bools = [(e[1], e[2]) for e in called_function_args_raw_information]
    return (
        recursively_visited,
        called_function_cu.name,
        called_function_args_raw_information_bools,
        function_raw_information_cache,
    )



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import cast

from sympy import Expr, Integer, Max


def convert_discopop_to_microbench_workload(discopop_per_iteration_workload: Expr, iteration_count: Expr) -> Expr:
    """Converts the estimated workload into the workload measurement used by Microbench.
    According to Formula:

    Wm = ((Wd/iD)-w0)/(iD*wi) with

    Wm = Microbench workload
    Wd = DiscoPoP workload
    iD = iteration count
    w0 = 13, initialization workload for inner loops
    wi = 14, observed workload added by a single iteration of the inner loop according to DiscoPoP's workload definition

    returns: Wm

    source: Bertins Thesis: TODO
    """
    # todo
    w0 = 13
    wi = 14

    # discopop_per_iteration_workload is equivalent to (Wd/iD)
    # for the same reason, iteration_count is disregarded
    Wm = cast(Expr, (discopop_per_iteration_workload - w0) / wi)
    # do not allow negative return values
    return Max(Wm, Integer(1))


def convert_microbench_to_discopop_workload(microbench_workload: Expr, iteration_count: Expr) -> Expr:
    """Converts the estimated workload into the workload measurement used by DiscoPoP.
    According to Formula:

    Wd = iD(Wm * iD * wi + w0) with

    Wm = Microbench workload
    Wd = DiscoPoP workload
    iD = iteration count
    w0 = 13, initialization workload for inner loops
    wi = 14, observed workload added by a single iteration of the inner loop according to DiscoPoP's workload definition

    returns: Wd

    source: Bertins Thesis: TODO
    """
    wi = Integer(14)
    w0 = Integer(13)

    return cast(Expr, iteration_count * (microbench_workload * iteration_count * wi + w0))



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os.path
import subprocess
from pathlib import Path
from typing import Dict

from discopop_library.PatchGenerator.PatchGeneratorArguments import PatchGeneratorArguments


def get_diffs_from_modified_code(
    file_mapping: Dict[int, Path], file_id_to_modified_code: Dict[int, str], arguments: PatchGeneratorArguments
) -> Dict[int, str]:
    patches: Dict[int, str] = dict()
    for file_id in file_id_to_modified_code:
        # get path to original code
        original_file_path = file_mapping[file_id]
        # create temporary modified code
        modified_file_path = original_file_path.parent / (original_file_path.name + ".discopop_patch_generator.temp")
        if arguments.verbose:
            print("Original: ", original_file_path)
            print("Modified:  ", modified_file_path)

        try:
            with open(modified_file_path, "w") as f:
                f.write(file_id_to_modified_code[file_id])
        except PermissionError:
            continue

        # calculate diff
        diff_name = original_file_path.parent / (original_file_path.name + ".discopop_patch_generator.diff")
        command = [
            "diff",
            "-Naru",
            original_file_path.as_posix(),
            modified_file_path.as_posix(),
        ]
        result = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=os.getcwd(),
        )
        if result.returncode != 0:
            if arguments.verbose:
                print("RESULT: ", result.returncode)
                print("STDERR:")
                print(result.stderr)
                print("STDOUT: ")
                print(result.stdout)

        # save diff
        patches[file_id] = result.stdout

        # cleanup environment
        if os.path.exists(modified_file_path):
            os.remove(modified_file_path)
        if os.path.exists(diff_name):
            os.remove(diff_name)

    return patches



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import tkinter
from tkinter import *
from typing import List, Tuple, Dict, Optional, cast

import networkx as nx  # type: ignore
from sympy import Symbol

from discopop_explorer.PEGraphX import PEGraphX
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.gui.plotting.CostModels import (
    plot_CostModels,
    plot_CostModels_using_function_path_selections,
    print_current_function_path_selections,
    print_current_substitutions,
    print_simplified_function,
)
from discopop_library.discopop_optimizer.gui.presentation.ChoiceDetails import (
    display_choices_for_model,
)
from discopop_library.discopop_optimizer.utilities.optimization.GlobalOptimization.RandomSamples import (
    find_quasi_optimal_using_random_samples,
)


def show_options(
    pet: PEGraphX,
    graph: nx.DiGraph,
    experiment: Experiment,
    options: List[Tuple[CostModel, ContextObject, str]],
    sorted_free_symbols: List[Symbol],
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    free_symbol_distributions: Dict[Symbol, FreeSymbolDistribution],
    function_root: FunctionRoot,
    parent_frame: tkinter.Frame,
    spawned_windows: List[tkinter.Toplevel],
    window_title=None,
) -> List[Tuple[CostModel, ContextObject, str]]:
    """Shows a tkinter table to browse and plot models"""
    # root = tkinter.Toplevel()
    # if window_title is not None:
    #    root.configure()
    #    root.title(window_title)
    root = tkinter.Toplevel(parent_frame.winfo_toplevel())
    spawned_windows.append(root)
    root.configure()
    root.title(window_title)

    rows = []
    column_headers = ["Label", "Decisions", "Options"]
    # set column headers
    header_cols = []
    for col_idx, header in enumerate(column_headers):
        e = Entry(root, relief=RIDGE)
        e.grid(row=0, column=col_idx, sticky=NSEW)
        e.insert(END, header)
        e.configure(state=DISABLED, disabledforeground="black")
        header_cols.append(e)
    rows.append(header_cols)

    label1 = Entry(root, relief=RIDGE)
    label1.grid(row=1, column=0, sticky=NSEW)
    label1.insert(END, "Current selection:")
    label1.configure(state=DISABLED, disabledforeground="black")

    label2 = Entry(root, relief=RIDGE)
    label2.grid(row=1, column=1, sticky=NSEW)
    label2.insert(END, str(experiment.selected_paths_per_function[function_root][0].path_decisions))
    label2.configure(state=DISABLED, disabledforeground="black")

    plot_button = Button(
        root,
        text="Plot using selections",
        command=lambda: plot_CostModels_using_function_path_selections(  # type: ignore
            experiment,
            [experiment.selected_paths_per_function[function_root][0]],  # type: ignore
            sorted_free_symbols,
            free_symbol_ranges,
            [function_root.name],
            title=function_root.name,
            super_title=function_root.name,
        ),
    )
    plot_button.grid(row=1, column=2, sticky=NSEW)

    print_selections_button = Button(
        root,
        text="Print selections",
        command=lambda: print_current_function_path_selections(  # type: ignore
            experiment,
        ),
    )
    print_selections_button.grid(row=1, column=3, sticky=NSEW)

    print_substitutions_button = Button(
        root,
        text="Print substitutions",
        command=lambda: print_current_substitutions(  # type: ignore
            experiment,
        ),
    )
    print_substitutions_button.grid(row=1, column=4, sticky=NSEW)

    print_simplified_function_button = Button(
        root,
        text="Print simplified",
        command=lambda: print_simplified_function(  # type: ignore
            experiment,
            [experiment.selected_paths_per_function[function_root][0]],  # type: ignore
            sorted_free_symbols,
            free_symbol_ranges,
            [function_root.name],
            title=function_root.name,
            super_title=function_root.name,
        ),
    )
    print_simplified_function_button.grid(row=1, column=5, sticky=NSEW)

    Button(
        root,
        text="Plot All",
        command=lambda: plot_CostModels(
            experiment,
            [t[0] for t in options],
            sorted_free_symbols,
            free_symbol_ranges,
            [t[2] for t in options],
            title="Full Plot",
            super_title=function_root.name,
        ),
    ).grid()  # type: ignore
    Button(
        root,
        text="Add Random (10)",
        command=lambda: add_random_models(
            root,
            pet,
            graph,
            experiment,
            [opt for opt in options if not opt[2].startswith("Rand ")],
            sorted_free_symbols,
            free_symbol_ranges,
            free_symbol_distributions,
            function_root,
            parent_frame,
            spawned_windows,
            window_title,
        ),
    ).grid()
    Button(root, text="Save Models", command=lambda: __save_models(experiment, function_root, options)).grid()
    Button(root, text="Close", command=lambda: root.destroy()).grid()

    # create option entries
    for row_idx, option_tuple in enumerate(options):
        option, context, option_name = option_tuple
        row_idx = row_idx + 1 + 5  # to account for column headers
        label = Entry(root, relief=RIDGE)
        label.grid(row=row_idx, column=0, sticky=NSEW)
        label.insert(END, option_name)
        label.configure(state=DISABLED, disabledforeground="black")

        decisions = Entry(root, relief=RIDGE)
        decisions.grid(row=row_idx, column=1, sticky=NSEW)
        decisions.insert(END, str(option.path_decisions))
        decisions.configure(state=DISABLED, disabledforeground="black")

        options_field = Entry(root, relief=RIDGE)
        options_field.grid(row=row_idx, column=2, sticky=NSEW)
        options_field.configure(state=DISABLED, disabledforeground="black")

        plot_button = Button(
            options_field,
            text="Plot",
            command=lambda opt=option, opt_name=option_name: plot_CostModels(  # type: ignore
                experiment,
                [opt],  # type: ignore
                sorted_free_symbols,
                free_symbol_ranges,
                [opt_name],
                title=opt_name,
                super_title=function_root.name,
            ),
        )
        plot_button.grid(row=0, column=0)
        details_button = Button(
            options_field,
            text="Details",
            command=lambda opt=option, opt_name=option_name: display_choices_for_model(  # type: ignore
                graph, opt, window_title=opt_name  # type: ignore
            ),
        )
        details_button.grid(row=0, column=1)

        export_code_button = Button(
            options_field,
            text="Export Code",
            command=lambda opt=option, opt_name=option_name, ctx=context: export_code(  # type: ignore
                pet, graph, experiment, opt, ctx, opt_name, function_root
            ),
        )
        export_code_button.grid(row=0, column=2)

        def __update_selection(cm, ctx):
            experiment.selected_paths_per_function[function_root] = (cm, ctx)
            experiment.substitutions[
                cast(Symbol, function_root.sequential_costs)
            ] = experiment.selected_paths_per_function[function_root][0].sequential_costs
            experiment.substitutions[
                cast(Symbol, function_root.parallelizable_costs)
            ] = experiment.selected_paths_per_function[function_root][0].parallelizable_costs
            # update displayed value
            label2.configure(state=NORMAL)
            label2.delete(0, END)
            label2.insert(0, str(cm.path_decisions))
            label2.configure(state=DISABLED)

        update_selection_button = Button(
            options_field,
            text="Update selection",
            command=lambda opt=option, opt_name=option_name, ctx=context: __update_selection(opt, ctx),  # type: ignore
        )
        update_selection_button.grid(row=0, column=3)

    root.mainloop()

    return options


def __save_models(
    experiment: Experiment,
    function_root: FunctionRoot,
    options: List[Tuple[CostModel, ContextObject, str]],
):
    print("SAVE: ", function_root)
    print("\ttype; ", type(function_root))
    experiment.function_models[function_root] = options
    print("Saved models for: ", function_root.name)


def add_random_models(
    root: Optional[tkinter.Toplevel],
    pet: PEGraphX,
    graph: nx.DiGraph,
    experiment: Experiment,
    conserve_options: List[Tuple[CostModel, ContextObject, str]],
    sorted_free_symbols: List[Symbol],
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    free_symbol_distributions: Dict[Symbol, FreeSymbolDistribution],
    function_root: FunctionRoot,
    parent_frame: Optional[tkinter.Frame],
    spawned_windows: List[tkinter.Toplevel],
    window_title=None,
    show_results: bool = True,
) -> List[Tuple[CostModel, ContextObject, str]]:
    if root is not None:
        # close window
        root.destroy()

    # generate random models
    # sort random models
    # add models to the list of options
    random_paths = 10  # amount of random models to be generated

    (
        minimum,
        maximum,
        median,
        lower_quartile,
        upper_quartile,
    ) = find_quasi_optimal_using_random_samples(
        experiment,
        graph,
        function_root,
        random_paths,
        sorted_free_symbols,
        free_symbol_ranges,
        free_symbol_distributions,
        verbose=True,
    )
    conserve_options.append((minimum[0], minimum[1], "Rand Minimum"))
    conserve_options.append((maximum[0], maximum[1], "Rand Maximum"))
    conserve_options.append((median[0], median[1], "Rand Median"))
    conserve_options.append((lower_quartile[0], lower_quartile[1], "Rand 25% Quartile"))
    conserve_options.append((upper_quartile[0], upper_quartile[1], "Rand 75% Quartile"))

    # plot
    if show_results:
        show_options(
            pet,
            graph,
            experiment,
            conserve_options,
            sorted_free_symbols,
            free_symbol_ranges,
            free_symbol_distributions,
            function_root,
            cast(tkinter.Frame, parent_frame),  # valid due to show_results flag
            spawned_windows,
            window_title,
        )

    return conserve_options



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from ast import Dict
from typing import List

import jsonpickle  # type: ignore

from discopop_explorer.PEGraphX import PEGraphX
from discopop_explorer.pattern_detectors.PatternBase import PatternBase
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.do_all_detector import DoAllInfo
from discopop_explorer.pattern_detectors.geometric_decomposition_detector import GDInfo
from discopop_explorer.pattern_detectors.pipeline_detector import PipelineInfo
from discopop_explorer.pattern_detectors.reduction_detector import ReductionInfo
from discopop_library.global_data.version.utils import get_version
from discopop_library.result_classes.PatternStorage import PatternStorage


class DetectionResult(object):
    version: str
    pet: PEGraphX
    patterns: PatternStorage

    def __init__(self, pet: PEGraphX):
        self.version = get_version()  # discopop version
        self.pet = pet
        self.patterns = PatternStorage()
        pass

    def __str__(self):
        result_str = ""
        for v in self.__dict__.values():
            if type(v) == PEGraphX:
                continue
            value_str = "\n\n\n"
            for entry in v:
                try:
                    value_str += str(entry) + "\n\n"
                except NotImplementedError:
                    value_str += entry.to_string(self.pet) + "\n\n"
            result_str += value_str
        return result_str

    def dump_to_pickled_json(self):
        """Encodes and returns the entire Object into a pickled json string.
        The encoded string can be reconstructed into an object by using:
        jsonpickle.decode(json_str)

        :return: encoded string
        """
        return jsonpickle.encode(self)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import tkinter as tk
from tkinter import filedialog
from tkinter import ttk

from DEPRECATED_discopop_wizard.screens.utils import create_tool_tip


def show_settings_screen(wizard):
    # clear content frame
    for c in wizard.window_frame.winfo_children():
        c.destroy()

    # build settings frame
    frame = tk.Frame(wizard.window_frame)
    frame.grid(row=1, column=1)

    # show labels
    tk.Label(frame, text="Directories:", justify=tk.RIGHT, font=wizard.style_font_bold).grid(
        row=1, column=1, sticky="ew"
    )
    tk.Label(frame, text="DiscoPoP build:", justify=tk.RIGHT, anchor="e").grid(row=3, column=1, sticky="ew")
    tk.Label(frame, text="go/bin directory:", justify=tk.RIGHT, anchor="e").grid(row=4, column=1, sticky="ew")

    ttk.Separator(frame, orient="horizontal").grid(row=5, column=1, sticky="ew", pady=10)
    tk.Label(frame, text="Executables:", justify=tk.RIGHT, font=wizard.style_font_bold).grid(row=6, column=1)
    tk.Label(frame, text="clang:", justify=tk.RIGHT, anchor="e").grid(row=7, column=1, sticky="ew")
    tk.Label(frame, text="clang++:", justify=tk.RIGHT, anchor="e").grid(row=8, column=1, sticky="ew")
    tk.Label(frame, text="llvm-ar:", justify=tk.RIGHT, anchor="e").grid(row=9, column=1, sticky="ew")
    tk.Label(frame, text="llvm-link:", justify=tk.RIGHT, anchor="e").grid(row=10, column=1, sticky="ew")
    tk.Label(frame, text="llvm-dis:", justify=tk.RIGHT, anchor="e").grid(row=11, column=1, sticky="ew")
    tk.Label(frame, text="llvm-opt:", justify=tk.RIGHT, anchor="e").grid(row=12, column=1, sticky="ew")
    tk.Label(frame, text="llvm-llc:", justify=tk.RIGHT, anchor="e").grid(row=13, column=1, sticky="ew")

    ttk.Separator(frame, orient="horizontal").grid(row=14, column=1, sticky="ew", pady=10)
    tk.Label(frame, text="Options:", justify=tk.RIGHT, font=wizard.style_font_bold).grid(row=15, column=1)
    tk.Label(frame, text="Use Docker Container for profiling:", justify=tk.RIGHT, anchor="e").grid(
        row=16, column=1, sticky="ew"
    )

    def __get_field_state():
        return tk.DISABLED if wizard.settings.use_docker_container_for_profiling else tk.NORMAL

    # show input fields
    discopop_build = tk.Entry(frame)
    discopop_build.grid(row=3, column=2, sticky="ew")
    discopop_build.insert(tk.END, wizard.settings.discopop_build_dir)
    discopop_build.config(state=__get_field_state())
    create_tool_tip(discopop_build, "Path to DiscoPoP build directory.")

    go_bin_path = tk.Entry(frame)
    go_bin_path.grid(row=4, column=2, sticky="ew")
    go_bin_path.insert(tk.END, wizard.settings.go_bin)
    go_bin_path.config(state=__get_field_state())
    create_tool_tip(go_bin_path, "Path to the bin folder inside the installation folder of Go.")

    clang = tk.Entry(frame, width=50)
    clang.grid(row=7, column=2, sticky="ew")
    clang.insert(tk.END, wizard.settings.clang)
    clang.config(state=__get_field_state())
    create_tool_tip(clang, "Path to the clang executable.")

    clangpp = tk.Entry(frame)
    clangpp.grid(row=8, column=2, sticky="ew")
    clangpp.insert(tk.END, wizard.settings.clangpp)
    clangpp.config(state=__get_field_state())
    create_tool_tip(clangpp, "Path to the clang++ executable.")

    llvm_ar = tk.Entry(frame)
    llvm_ar.grid(row=9, column=2, sticky="ew")
    llvm_ar.insert(tk.END, wizard.settings.llvm_ar)
    llvm_ar.config(state=__get_field_state())
    create_tool_tip(llvm_ar, "Path to the llvm-ar executable.")

    llvm_link = tk.Entry(frame)
    llvm_link.grid(row=10, column=2, sticky="ew")
    llvm_link.insert(tk.END, wizard.settings.llvm_link)
    llvm_link.config(state=__get_field_state())
    create_tool_tip(llvm_link, "Path to the llvm_link executable.")

    llvm_dis = tk.Entry(frame)
    llvm_dis.grid(row=11, column=2, sticky="ew")
    llvm_dis.insert(tk.END, wizard.settings.llvm_dis)
    llvm_dis.config(state=__get_field_state())
    create_tool_tip(llvm_dis, "Path to the llvm_dis executable.")

    llvm_opt = tk.Entry(frame)
    llvm_opt.grid(row=12, column=2, sticky="ew")
    llvm_opt.insert(tk.END, wizard.settings.llvm_opt)
    llvm_opt.config(state=__get_field_state())
    create_tool_tip(llvm_opt, "Path to the llvm_opt executable.")

    llvm_llc = tk.Entry(frame)
    llvm_llc.grid(row=13, column=2, sticky="ew")
    llvm_llc.insert(tk.END, wizard.settings.llvm_llc)
    llvm_llc.config(state=__get_field_state())
    create_tool_tip(llvm_llc, "Path to the llvm_llc executable.")

    use_docker_container_var = tk.IntVar(value=1 if wizard.settings.use_docker_container_for_profiling else 0)
    use_docker_container = tk.Checkbutton(frame, variable=use_docker_container_var)
    create_tool_tip(
        use_docker_container,
        "When un-checking, please save and re-open" " the settings to enable input fields!",
    )

    use_docker_container.grid(row=16, column=2)

    # show path selector buttons
    tk.Button(frame, text="Select", command=lambda: __overwrite_with_selection(discopop_build)).grid(row=3, column=3)
    tk.Button(frame, text="Select", command=lambda: __overwrite_with_selection(go_bin_path)).grid(row=4, column=3)
    tk.Button(frame, text="Select", command=lambda: __overwrite_with_selection(clang)).grid(row=7, column=3)
    tk.Button(frame, text="Select", command=lambda: __overwrite_with_selection(clangpp)).grid(row=8, column=3)
    tk.Button(frame, text="Select", command=lambda: __overwrite_with_selection(llvm_ar)).grid(row=9, column=3)
    tk.Button(frame, text="Select", command=lambda: __overwrite_with_selection(llvm_link)).grid(row=10, column=3)
    tk.Button(frame, text="Select", command=lambda: __overwrite_with_selection(llvm_dis)).grid(row=11, column=3)
    tk.Button(frame, text="Select", command=lambda: __overwrite_with_selection(llvm_opt)).grid(row=12, column=3)
    tk.Button(frame, text="Select", command=lambda: __overwrite_with_selection(llvm_llc)).grid(row=13, column=3)

    # show save button
    tk.Button(
        frame,
        text="Save",
        command=lambda: save_settings(
            wizard,
            discopop_build,
            go_bin_path,
            clang,
            clangpp,
            llvm_ar,
            llvm_link,
            llvm_dis,
            llvm_opt,
            llvm_llc,
            use_docker_container_var,
        ),
    ).grid(row=17, column=2, pady=10)


def __overwrite_with_selection(target: tk.Entry):
    prompt_result = tk.filedialog.askdirectory()
    if len(prompt_result) != 0:
        target.delete(0, tk.END)
        target.insert(0, prompt_result)


def save_settings(
    wizard,
    discopop_build: tk.Entry,
    go_bin_path: tk.Entry,
    clang: tk.Entry,
    clangpp: tk.Entry,
    llvm_ar: tk.Entry,
    llvm_link: tk.Entry,
    llvm_dis: tk.Entry,
    llvm_opt: tk.Entry,
    llvm_llc: tk.Entry,
    use_docker_container_var,
):
    wizard.settings.discopop_build_dir = discopop_build.get()
    wizard.settings.go_bin = go_bin_path.get()
    wizard.settings.clang = clang.get()
    wizard.settings.clangpp = clangpp.get()
    wizard.settings.llvm_ar = llvm_ar.get()
    wizard.settings.llvm_link = llvm_link.get()
    wizard.settings.llvm_dis = llvm_dis.get()
    wizard.settings.llvm_opt = llvm_opt.get()
    wizard.settings.llvm_llc = llvm_llc.get()
    wizard.settings.use_docker_container_for_profiling = True if use_docker_container_var.get() == 1 else False

    wizard.settings.save_to_file(config_dir=wizard.config_dir)

    # return to main screen
    wizard.show_main_screen()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import Optional, List


class Line(object):
    line_num: Optional[int]
    content: str
    owns_region: Optional[int]
    belongs_to_regions: List[int]
    belongs_to_original_line: int

    def __init__(self, parent_line_num: int, line_num=None, content=""):
        self.line_num = line_num
        self.content = content
        self.owns_region = None
        self.belongs_to_regions = []
        self.belongs_to_original_line = parent_line_num

    def get_indentation(self) -> str:
        return self.content[: len(self.content) - len(self.content.lstrip())]



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os.path
import shutil
from pathlib import Path
from typing import Dict

from discopop_library.PatchApplicator.PatchApplicatorArguments import PatchApplicatorArguments
from discopop_library.PatchApplicator.rollback import rollback_patches


def clear_patches(
    file_mapping: Dict[int, Path],
    arguments: PatchApplicatorArguments,
    applied_suggestions_file: str,
    patch_generator_dir: str,
) -> int:
    # save the currently applied suggestions. Overwrites old saves
    shutil.copyfile(applied_suggestions_file, applied_suggestions_file + ".save")

    # get the list of applied suggestions
    with open(applied_suggestions_file, "r") as f:
        applied_suggestions = json.loads(f.read())["applied"]

    # rollback all suggestions in inverse order
    applied_suggestions.reverse()
    retval = rollback_patches(
        applied_suggestions, file_mapping, arguments, applied_suggestions_file, patch_generator_dir
    )

    return retval



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os.path
import shutil
from typing import Dict

from discopop_library.CodeGenerator.CodeGenerator import from_json_strings
from discopop_library.JSONHandler.JSONHandler import read_patterns_from_json_to_json
from discopop_library.PatchGenerator.PatchGeneratorArguments import PatchGeneratorArguments
from discopop_library.PatchGenerator.from_optimizer_output import from_optimizer_output
from discopop_library.PatchGenerator.diffs import get_diffs_from_modified_code
from discopop_library.PatchGenerator.from_json_patterns import from_json_patterns
from discopop_library.PathManagement.PathManagement import load_file_mapping


def run(arguments: PatchGeneratorArguments):
    if arguments.verbose:
        print("Started DiscoPoP Patch Generator...")
    if arguments.verbose:
        print("Creating patch_generator directory...")
    patch_generator_dir = os.path.join(os.getcwd(), "patch_generator")
    if not os.path.exists(patch_generator_dir):
        os.mkdir(patch_generator_dir)

    # for compatibility reasons, initialize the file to store applied patches if it doesn't exist already
    # create a directory for the patch applicator
    patch_applicator_dir = os.path.join(os.getcwd(), "patch_applicator")
    if not os.path.exists(patch_applicator_dir):
        if arguments.verbose:
            print("Creating patch_applicator directory...")
        os.mkdir(patch_applicator_dir)
    # create a file to store applied suggestions
    applied_suggestions_file = os.path.join(patch_applicator_dir, "applied_suggestions.json")
    if not os.path.exists(applied_suggestions_file):
        if arguments.verbose:
            print("Creating applied_suggestions.json file...")
        with open(applied_suggestions_file, "w+") as f:
            f.write(json.dumps({"applied": []}))

    # get pattern file to load
    if arguments.add_from_json != "None":
        pattern_file_path = arguments.add_from_json
    else:
        pattern_file_path = os.path.join(os.getcwd(), "explorer", "patterns.json")
    if not os.path.exists(pattern_file_path):
        raise FileNotFoundError(
            "No pattern file found. Please execute the discopop_explorer in advance."
            + "\nExpected pattern file: "
            + pattern_file_path
        )
    file_mapping_path = os.path.join(os.getcwd(), "FileMapping.txt")
    if not os.path.exists(file_mapping_path):
        raise FileNotFoundError(
            "No file mapping found. Please execute the discopop_explorer in advance."
            + "\nExpected file: "
            + file_mapping_path
        )

    if arguments.verbose:
        print("Loading file mapping...")
    file_mapping = load_file_mapping(file_mapping_path)

    if arguments.verbose:
        print("Loading patterns...")
    patterns_by_type = read_patterns_from_json_to_json(pattern_file_path, [])
    if arguments.verbose:
        print("Patterns: ", patterns_by_type)

    from_json_patterns(arguments, patterns_by_type, file_mapping, patch_generator_dir)

    if arguments.verbose:
        print("Done.")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from discopop_library.discopop_optimizer.classes.edges.GenericEdge import GenericEdge
from discopop_library.discopop_optimizer.classes.edges.SuccessorEdge import SuccessorEdge


class TemporaryEdge(GenericEdge):
    pass

    def convert_to_successor_edge(self) -> SuccessorEdge:
        return SuccessorEdge()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os
from typing import Set, Dict, cast, Optional, List, Tuple

from discopop_explorer.PEGraphX import PEGraphX, NodeID, MemoryRegion
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Aliases import (
    VarName,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Dependency import Dependency
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Enums import UpdateType
from discopop_library.MemoryRegions.utils import get_sizes_of_memory_regions


class Update(object):
    synchronous_source_cu_id: NodeID
    asynchronous_source_cu_id: Optional[NodeID]
    sink_cu_id: NodeID
    memory_regions: Set[MemoryRegion]
    variable_names: Set[VarName]
    update_type: UpdateType
    last_write_locations: Dict[MemoryRegion, NodeID]
    asynchronous_possible: bool
    dependencies: Set[Dependency]

    def __init__(
        self,
        cu_id: NodeID,
        sink_cu_id: NodeID,
        memory_regions: Set[MemoryRegion],
        update_type: UpdateType,
        last_write_locations: Dict[MemoryRegion, NodeID],
    ):
        self.synchronous_source_cu_id = cu_id
        self.asynchronous_source_cu_id = None
        self.sink_cu_id = sink_cu_id
        self.memory_regions = memory_regions
        self.variable_names = set()
        self.update_type = update_type
        self.last_write_locations = last_write_locations
        self.dependencies = set()
        self.asynchronous_possible = False

    # Preserve, as it will be added in the future
    #        if False:  # disable asynchronous updates
    #            # check if an asynchronous update is possible
    #            # only allow asynchronous updates to the device in order to not get problems with
    #            # enforcing dependencies on the host --> tasks etc. would be required to implement the waiting behavior
    #            if self.update_type != UpdateType.FROM_DEVICE:
    #                import sys
    #
    #                print("ORIGINAL SOURCE: ", self.synchronous_source_cu_id, file=sys.stderr)
    #                self.asynchronous_possible = True
    #                asynchronous_source: Optional[NodeID] = None
    #                for mem_reg in self.memory_regions:
    #                    if self.last_write_locations[mem_reg] == self.synchronous_source_cu_id:
    #                        self.asynchronous_possible = False
    #                        break
    #                    if asynchronous_source is None:
    #                        asynchronous_source = self.last_write_locations[mem_reg]
    #                    if asynchronous_source != self.last_write_locations[mem_reg]:
    #                        self.asynchronous_possible = False
    #                        break
    #
    #                if self.asynchronous_possible:
    #                    # update the asynchronous source cu
    #                    self.asynchronous_source_cu_id = cast(NodeID, asynchronous_source)
    #
    #                    # create a dependency to ensure correctness
    #                    self.dependencies.add(
    #                        Dependency(
    #                            self.asynchronous_source_cu_id,
    #                            self.sink_cu_id,
    #                            self.variable_names,
    #                            self.memory_regions,
    #                        )
    #                    )

    def __str__(self):
        result_str = ""
        result_str += (
            str(self.update_type)
            + " @ "
            + self.synchronous_source_cu_id
            + " -> "
            + self.sink_cu_id
            + " : "
            + str(self.variable_names)
            + "/"
            + str(self.memory_regions)
            + " last_writes@ "
            + str(self.last_write_locations)
            + " async "
            if self.asynchronous_possible
            else ""
        )
        return result_str

    def __eq__(self, other):
        if (
            self.synchronous_source_cu_id,
            self.asynchronous_source_cu_id,
            self.sink_cu_id,
            #            tuple(self.memory_regions),  # leads to duplicated outputs
            tuple(self.variable_names),
            self.update_type,
            tuple(self.last_write_locations.items()),
            self.asynchronous_possible,
            tuple(self.dependencies),
        ) == (
            other.synchronous_source_cu_id,
            other.asynchronous_source_cu_id,
            other.sink_cu_id,
            #            tuple(other.memory_regions),  # leads to duplicated outputs
            tuple(other.variable_names),
            other.update_type,
            tuple(other.last_write_locations.items()),
            other.asynchronous_possible,
            tuple(other.dependencies),
        ):
            return True
        return False

    def __hash__(self):
        return hash(
            (
                self.synchronous_source_cu_id,
                self.asynchronous_source_cu_id,
                self.sink_cu_id,
                tuple(self.memory_regions),
                tuple(self.variable_names),
                self.update_type,
                tuple(self.last_write_locations.items()),
                self.asynchronous_possible,
                tuple(self.dependencies),
            )
        )

    def get_position_identifier(self):
        # used to join multiple elements
        if self.asynchronous_possible:
            return (self.sink_cu_id, self.asynchronous_source_cu_id, self.update_type)
        else:
            return (self.sink_cu_id, self.synchronous_source_cu_id, self.update_type)

    def join(self, other):
        self.variable_names.update(other.variable_names)
        self.memory_regions.update(other.memory_regions)
        self.dependencies.update(other.dependencies)

    def get_as_metadata_using_memory_regions(self, pet: PEGraphX):
        return [
            self.synchronous_source_cu_id,
            self.sink_cu_id,
            self.update_type,
            str(self.memory_regions),
            pet.node_at(self.synchronous_source_cu_id).end_position(),
        ]

    def get_as_metadata_using_variable_names(self, pet: PEGraphX, project_folder_path: str):
        # get type of mapped variables
        var_names_types_and_sizes: List[Tuple[VarName, str, int]] = []
        for var_name in self.variable_names:
            var_obj = pet.get_variable(self.sink_cu_id, var_name)
            source_cu_id = (
                self.asynchronous_source_cu_id if self.asynchronous_possible else self.synchronous_source_cu_id
            )
            if var_obj is None:
                var_obj = pet.get_variable(cast(NodeID, source_cu_id), var_name)
            if var_obj is None:
                var_names_types_and_sizes.append((var_name, "", 1))
            else:
                var_names_types_and_sizes.append((var_name, var_obj.type, var_obj.sizeInByte))
        # add [..] to variable name if required (type contains "**")

        # get size of memory region
        memory_region_sizes = get_sizes_of_memory_regions(
            self.memory_regions, os.path.join(project_folder_path, "profiler/memory_regions.txt")
        )
        if len(memory_region_sizes) > 0:
            max_mem_reg_size = max(memory_region_sizes.values())
            # divide memory region size by size of variable
            # construct new list of modified var names
            modified_var_names = [
                (
                    vn + "[:]" if "**" in t else vn
                )  # (vn + "[:" + str(int(max_mem_reg_size / s)) + "]" if "**" in t else vn)
                for vn, t, s in var_names_types_and_sizes
            ]
        else:
            modified_var_names = [(vn + "[:]" if "**" in t else vn) for vn, t, s in var_names_types_and_sizes]

        # determine update position in code
        # todo consider asynchronous updates
        if self.update_type == UpdateType.FROM_DEVICE:
            # data required by host. Perform update before the sink CU
            update_position = pet.node_at(self.sink_cu_id).start_position()
        elif self.update_type == UpdateType.TO_DEVICE:
            # data required by device. Perform update before the sink CU
            update_position = pet.node_at(self.sink_cu_id).start_position()
        elif self.update_type == UpdateType.TO_FROM_DEVICE:
            # synchronize inbetween source and sink -> before the start of sink CU
            update_position = pet.node_at(self.sink_cu_id).start_position()
        elif self.update_type == UpdateType.ALLOCATE:
            update_position = pet.node_at(self.sink_cu_id).start_position()
        else:
            # updating inbetween both CUs should be a safe fallback
            update_position = pet.node_at(self.sink_cu_id).start_position()

        return [
            self.synchronous_source_cu_id,
            self.sink_cu_id,
            self.update_type,
            ",".join(modified_var_names),
            update_position,
        ]

    def get_as_metadata_using_variable_names_and_memory_regions(self, pet: PEGraphX, project_folder_path: str):
        # get type of mapped variables
        var_names_types_and_sizes: List[Tuple[VarName, str, int]] = []
        for var_name in self.variable_names:
            var_obj = pet.get_variable(self.sink_cu_id, var_name)
            source_cu_id = (
                self.asynchronous_source_cu_id if self.asynchronous_possible else self.synchronous_source_cu_id
            )
            if var_obj is None:
                var_obj = pet.get_variable(cast(NodeID, source_cu_id), var_name)
            if var_obj is None:
                var_names_types_and_sizes.append((var_name, "", 1))
            else:
                var_names_types_and_sizes.append((var_name, var_obj.type, var_obj.sizeInByte))
        # add [..] to variable name if required (type contains "**")

        # get size of memory region
        memory_region_sizes = get_sizes_of_memory_regions(
            self.memory_regions, os.path.join(project_folder_path, "profiler/memory_regions.txt")
        )
        if len(memory_region_sizes) > 0:
            max_mem_reg_size = max(memory_region_sizes.values())
            # divide memory region size by size of variable
            # construct new list of modified var names
            modified_var_names = [
                (
                    vn + "[:]" if "**" in t else vn
                )  # (vn + "[:" + str(int(max_mem_reg_size / s)) + "]" if "**" in t else vn)
                for vn, t, s in var_names_types_and_sizes
            ]
        else:
            modified_var_names = [(vn + "[:]" if "**" in t else vn) for vn, t, s in var_names_types_and_sizes]

        return [
            self.synchronous_source_cu_id,
            self.sink_cu_id,
            self.update_type,
            str(modified_var_names) + "/" + str(self.memory_regions),
            pet.node_at(self.synchronous_source_cu_id).end_position(),
        ]

    def convert_memory_regions_to_variable_names(
        self,
        pet: PEGraphX,
        memory_regions_to_functions_and_variables: Dict[MemoryRegion, Dict[NodeID, Set[VarName]]],
    ):
        self.variable_names = set()
        parent_function_id = pet.get_parent_function(pet.node_at(self.synchronous_source_cu_id)).id

        for mem_reg in self.memory_regions:
            if parent_function_id in memory_regions_to_functions_and_variables[mem_reg]:
                self.variable_names.update(memory_regions_to_functions_and_variables[mem_reg][parent_function_id])
            elif self.synchronous_source_cu_id in memory_regions_to_functions_and_variables[mem_reg]:
                self.variable_names.update(
                    memory_regions_to_functions_and_variables[mem_reg][self.synchronous_source_cu_id]
                )
            elif self.sink_cu_id in memory_regions_to_functions_and_variables[mem_reg]:
                self.variable_names.update(memory_regions_to_functions_and_variables[mem_reg][self.sink_cu_id])
            else:
                self.variable_names.add(VarName("UNDETERMINED(" + mem_reg + ")"))

    def remove_dummy_marks(self):
        if self.sink_cu_id.startswith("dummy:"):
            self.sink_cu_id = NodeID(self.sink_cu_id[6:])
        if self.synchronous_source_cu_id.startswith("dummy:"):
            self.synchronous_source_cu_id = NodeID(self.synchronous_source_cu_id[6:])



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, cast, Set
from discopop_explorer.PEGraphX import EdgeType, NodeID
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    get_all_function_nodes,
    get_children,
    get_in_mutex_edges,
    get_in_options,
    get_out_mutex_edges,
    get_out_options,
    get_requirements,
    get_successors,
)
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at


def check_configuration_validity(
    experiment: Experiment, arguments: OptimizerArguments, configuration: List[int]
) -> bool:
    """Returns True if the given configuration is valid. Returns False otherwise."""

    # check requirements edges
    for node_id in configuration:
        requirements = get_requirements(experiment.optimization_graph, node_id)
        for r in requirements:
            if r not in configuration:
                # requirement not satisfied
                return False
    # check option edges (for mutual exclusivity)
    for node_id in configuration:
        mutex_options = get_out_mutex_edges(experiment.optimization_graph, node_id)
        if len([e for e in configuration if e in mutex_options]) != 0:
            # mutual exclusivity of suggestions violated
            return False
    # check for nested parallelism
    if not arguments.allow_nested_parallelism:
        if __nested_parallelism_found(experiment, configuration, arguments):
            return False
    # check for duplicated suggestion ids
    # note: this check should not be necessary and might be removed in the future
    applied_suggestion_ids: Set[int] = set()
    for node_id in configuration:
        if node_id in experiment.node_id_to_suggestion_dict:
            if experiment.node_id_to_suggestion_dict[node_id] in applied_suggestion_ids:
                return False
            applied_suggestion_ids.add(experiment.node_id_to_suggestion_dict[node_id])

    return True


def __nested_parallelism_found(experiment: Experiment, configuration: List[int], arguments: OptimizerArguments) -> bool:
    """checks for nested parallelism in the given configuration.
    Returns True if the configuration leads to nested parallelism.
    Returns False otherwise."""
    visited_nodes = []
    children_queue = []
    # initialize knowledge
    all_function_nodes = [
        data_at(experiment.optimization_graph, nid) for nid in get_all_function_nodes(experiment.optimization_graph)
    ]

    # initialize_queue
    for node_id in configuration:
        node = data_at(experiment.optimization_graph, node_id)
        if not node.represents_sequential_version():
            children_queue += __filter_for_relevant_options(
                experiment, configuration, get_children(experiment.optimization_graph, node_id)
            )

    while children_queue:
        current_node_id = children_queue.pop(0)
        current_node = data_at(experiment.optimization_graph, current_node_id)
        # check for nested parallelism
        if not current_node.represents_sequential_version():
            return True

        # add successors and children
        visited_nodes.append(current_node_id)
        children_queue += __filter_for_relevant_options(
            experiment,
            configuration,
            [
                sid
                for sid in get_successors(experiment.optimization_graph, current_node_id)
                if sid not in visited_nodes and sid not in children_queue
            ],
        )

        children_queue += __filter_for_relevant_options(
            experiment,
            configuration,
            [
                cid
                for cid in get_children(experiment.optimization_graph, current_node_id)
                if cid not in visited_nodes and cid not in children_queue
            ],
        )
        if arguments.check_called_function_for_nested_parallelism:
            # add called functions
            called_cu_ids: List[str] = [
                str(t)
                for s, t, d in experiment.detection_result.pet.out_edges(
                    cast(NodeID, current_node.original_cu_id), EdgeType.CALLSNODE
                )
            ]
            # filter for called FunctionRoots
            called_function_nodes = [fr for fr in all_function_nodes if str(fr.original_cu_id) in called_cu_ids]

            # remove duplicates
            called_function_nodes = list(set(called_function_nodes))
            # add to children_queue
            children_queue += [
                fid.node_id
                for fid in called_function_nodes
                if fid.node_id not in visited_nodes and fid.node_id not in children_queue
            ]

    return False


def __filter_for_relevant_options(
    experiment: Experiment, configuration: List[int], to_be_cleaned: List[int]
) -> List[int]:
    """removes all node_ids from the list to_be_cleaned which are not relevant for the given configuration (i.e. parallelization opportunities which are not chosen.)"""
    cleaned_list: List[int] = []
    for node_id in to_be_cleaned:
        if node_id in configuration:
            cleaned_list.append(node_id)
            continue
        options = get_out_mutex_edges(experiment.optimization_graph, node_id) + get_in_mutex_edges(
            experiment.optimization_graph, node_id
        )
        if len(options) == 0:
            cleaned_list.append(node_id)
            continue
        else:
            # node_id is not in configuration and options exist, i.e. it is not selected by the configuration and has to be ignored
            continue
    return cleaned_list



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import subprocess
import tkinter as tk
from typing import Optional, cast

from DEPRECATED_discopop_wizard.classes.ProfilingContainer import ProfilingContainer
from DEPRECATED_discopop_wizard.screens.suggestions.overview import show_suggestions_overview_screen


class ExecutionView(object):
    def __init__(
        self,
        execution_configuration,
        wizard,
        details_frame: Optional[tk.Frame],
        headless_mode: bool = False,
    ):
        self.execution_configuration = execution_configuration
        self.wizard = wizard
        self.details_frame = details_frame
        self.headless_mode = headless_mode
        self.__execute()

    def __execute(self):
        # prepare environment
        if self.wizard.settings.use_docker_container_for_profiling:
            # start container if not already present. Required when enabling container usage after start of application.
            if self.wizard.profiling_container is None:
                self.wizard.profiling_container = ProfilingContainer(self.wizard)
            self.wizard.profiling_container.analyze_project(self)
            # todo add display of suggestions
            if not self.headless_mode:
                self.__show_suggestions()
        else:
            # prepare command
            command = self.__assemble_command_string()
            # execute command
            return_code = self.__execute_command(command)
            if return_code == 0:
                # show suggestions, stored in project_path/patterns.txt
                if not self.headless_mode:
                    self.__show_suggestions()

    def __assemble_command_string(self) -> str:
        # assemble command for regular execution
        command = ""
        # settings
        command = self.wizard.settings.discopop_build_dir + "/scripts/runDiscoPoP "
        command += '--llvm-clang "' + self.wizard.settings.clang + '" '
        command += '--llvm-clang++ "' + self.wizard.settings.clangpp + '" '
        command += '--llvm-ar "' + self.wizard.settings.llvm_ar + '" '
        command += '--llvm-link "' + self.wizard.settings.llvm_link + '" '
        command += '--llvm-dis "' + self.wizard.settings.llvm_dis + '" '
        command += '--llvm-opt "' + self.wizard.settings.llvm_opt + '" '
        command += '--llvm-llc "' + self.wizard.settings.llvm_llc + '" '
        command += '--gllvm "' + self.wizard.settings.go_bin + '" '
        # execution configuration
        command += '--project "' + self.execution_configuration.value_dict["project_path"] + '" '
        command += '--linker-flags "' + self.execution_configuration.value_dict["linker_flags"] + '" '
        command += '--executable-name "' + self.execution_configuration.value_dict["executable_name"] + '" '
        command += '--executable-arguments "' + self.execution_configuration.value_dict["executable_arguments"] + '" '
        command += '--make-flags "' + self.execution_configuration.value_dict["make_flags"] + '" '
        command += '--make-target "' + self.execution_configuration.value_dict["make_target"] + '" '
        command += (
            "--memory-profiling-skip-function-arguments "
            if self.execution_configuration.value_dict["memory_profiling_skip_function_parameters"] == 1
            else ""
        )
        command += '--explorer-flags "' + self.execution_configuration.value_dict["explorer_flags"] + '" '

        return command

    def __execute_command(self, command: str) -> int:
        with subprocess.Popen(command, stdout=subprocess.PIPE, bufsize=1, universal_newlines=True, shell=True) as p:
            if p.stdout is None:
                print("command execution was not successfull")
            else:
                for line in p.stdout:
                    line = line.replace("\n", "")
                    self.__print_to_console(line)
                    if not self.headless_mode:
                        self.wizard.console.print(line)
        if p.returncode != 0:
            self.__print_to_console("An error occurred during the execution!")  # Error message
            if not self.headless_mode:
                self.wizard.console.print("An error occurred during the execution!")
            for line in str(subprocess.CalledProcessError(p.returncode, p.args)).split("\n"):
                line = line.replace("\n", "")
                self.__print_to_console(line)
                if not self.headless_mode:
                    self.wizard.console.print(line)
        return p.returncode

    def __print_to_console(self, msg: str):
        print(msg)

    def __show_suggestions(self):
        show_suggestions_overview_screen(self.wizard, cast(tk.Frame, self.details_frame), self.execution_configuration)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os.path
from dataclasses import dataclass
from typing import List, Optional


@dataclass
class OptimizerArguments(object):
    """Container Class for the arguments passed to the discopop_optimizer"""

    verbose: bool
    interactive: bool
    exhaustive: bool
    evolutionary: Optional[List[str]]
    doall_microbench_file: str
    reduction_microbench_file: str
    allow_nested_parallelism: bool
    plot: bool
    system_configuration_path: str
    check_called_function_for_nested_parallelism: bool

    def __post_init__(self):
        # fix correct optimization method
        if not self.exhaustive:
            if self.evolutionary == None:
                self.evolutionary = [str(50), str(5)]
        elif self.evolutionary != None:
            self.evolutionary = None

        self.__validate()

    def __validate(self):
        """Validate the arguments passed to the discopop_optimizer, e.g check if given files exist"""
        if self.doall_microbench_file is not "None":
            if not os.path.isfile(self.doall_microbench_file):
                raise FileNotFoundError(f"Microbenchmark file not found: {self.doall_microbench_file}")
        if self.reduction_microbench_file is not "None":
            if not os.path.isfile(self.reduction_microbench_file):
                raise FileNotFoundError(f"Microbenchmark file not found: {self.reduction_microbench_file}")
        pass



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os.path
from typing import Dict


def load_line_mapping(discopop_path: str = "") -> Dict[str, Dict[str, int]]:
    """Loads and returns the line_mapping dictionary"""
    line_mapping_dict: Dict[str, Dict[str, int]] = dict()
    with open(os.path.join(discopop_path, "line_mapping.json")) as f:
        line_mapping_dict = json.load(f)
    return line_mapping_dict



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os.path
from typing import Set, Tuple, List

from discopop_explorer.PEGraphX import PEGraphX, NodeID, MemoryRegion
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Aliases import (
    VarName,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Dependency import Dependency
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Enums import (
    EntryPointType,
    EntryPointPositioning,
)
from discopop_library.MemoryRegions.utils import (
    get_sizes_of_memory_regions,
)


class EntryPoint(object):
    var_names: Set[VarName]
    memory_regions: Set[MemoryRegion]
    source_cu_id: NodeID
    sink_cu_id: NodeID
    entry_point_type: EntryPointType
    pragma_line: str
    entry_point_positioning: EntryPointPositioning
    dependencies: Set[Dependency]

    def __init__(
        self,
        pet: PEGraphX,
        var_names: Set[VarName],
        memory_regions: Set[MemoryRegion],
        source_cu_id: NodeID,
        sink_cu_id: NodeID,
        entry_point_type: EntryPointType,
    ):
        self.var_names = var_names
        self.memory_regions = memory_regions
        self.source_cu_id = source_cu_id
        self.sink_cu_id = sink_cu_id

        # todo handle async
        self.entry_point_type = entry_point_type
        self.pragma_line = pet.node_at(sink_cu_id).start_position()
        self.entry_point_positioning = EntryPointPositioning.BEFORE_CU
        self.dependencies = set()

    def __str__(self):
        return (
            "EntryPoint("
            + str(self.var_names)
            + ", "
            + str(self.memory_regions)
            + ", @ "
            + self.source_cu_id
            + " -> "
            + self.sink_cu_id
            + " : "
            + str(self.entry_point_type)
            + ")"
        )

    def __eq__(self, other):
        if (
            tuple(self.var_names),
            #            tuple(self.memory_regions),  # leads to duplicated outputs
            self.source_cu_id,
            self.sink_cu_id,
            self.entry_point_type,
            self.pragma_line,
            self.entry_point_positioning,
            tuple(self.dependencies),
        ) == (
            tuple(other.var_names),
            #            tuple(other.memory_regions),  # leads to duplicated outputs
            other.source_cu_id,
            other.sink_cu_id,
            other.entry_point_type,
            other.pragma_line,
            other.entry_point_positioning,
            tuple(other.dependencies),
        ):
            return True
        return False

    def __hash__(self):
        return hash(
            (
                tuple(self.var_names),
                tuple(self.memory_regions),
                self.source_cu_id,
                self.sink_cu_id,
                self.entry_point_type,
                self.pragma_line,
                self.entry_point_positioning,
                tuple(self.dependencies),
            )
        )

    def get_position_identifier(self):
        # used to join multiple elements
        return (
            self.sink_cu_id,
            self.source_cu_id,
            self.entry_point_type,
            self.entry_point_positioning,
        )

    def join(self, other):
        self.var_names.update(other.var_names)
        self.memory_regions.update(other.memory_regions)
        self.dependencies.update(other.dependencies)

    def get_as_metadata(self, pet: PEGraphX, project_folder_path: str):
        # get type of mapped variables
        var_names_types_and_sizes: List[Tuple[VarName, str, int]] = []
        for var_name in self.var_names:
            var_obj = pet.get_variable(self.sink_cu_id, var_name)
            if var_obj is None:
                var_obj = pet.get_variable(self.source_cu_id, var_name)
            if var_obj is None:
                var_names_types_and_sizes.append((var_name, "", 1))
            else:
                var_names_types_and_sizes.append((var_name, var_obj.type, var_obj.sizeInByte))
        # add [..] to variable name if required (type contains "**")

        # get size of memory region
        memory_region_sizes = get_sizes_of_memory_regions(
            self.memory_regions, os.path.join(project_folder_path, "profiler/memory_regions.txt")
        )
        if len(memory_region_sizes) > 0:
            max_mem_reg_size = max(memory_region_sizes.values())
            # divide memory region size by size of variable
            # construct new list of modified var names
            modified_var_names = [
                (
                    vn + "[:]" if "**" in t else vn
                )  # (vn + "[:" + str(int(max_mem_reg_size / s)) + "]" if "**" in t else vn)
                for vn, t, s in var_names_types_and_sizes
            ]
        else:
            modified_var_names = [(vn + "[:]" if "**" in t else vn) for vn, t, s in var_names_types_and_sizes]

        return [
            ",".join(modified_var_names),
            self.source_cu_id,
            self.sink_cu_id,
            self.entry_point_type,
            self.pragma_line,
            self.entry_point_positioning,
        ]



import os
import pathlib
import unittest

import jsonpickle

from discopop_library.result_classes.DetectionResult import DetectionResult
from test.utils.validator_classes.DoAllInfoForValidation import DoAllInfoForValidation


class TestMethods(unittest.TestCase):
    def test(self):
        current_dir = pathlib.Path(__file__).parent.resolve()
        discopop_dir = os.path.join(current_dir, "..", "..", "..")
        build_dir = os.path.join(discopop_dir, "build")

        src_dir = os.path.join(current_dir, "src")
        os.chdir(src_dir)

        # create FileMapping
        os.system("" + os.path.join(build_dir, "scripts", "dp-fmap"))

        # build
        # make_command = "DP_FM_PATH=" + os.path.join(src_dir, "FileMapping.txt") + " "
        make_command = ""
        make_command += "CC=" + os.path.join(build_dir, "scripts", "CC_wrapper.sh") + " "
        make_command += "CXX=" + os.path.join(build_dir, "scripts", "CXX_wrapper.sh") + " "
        # make_command += "LINKER=" + os.path.join(build_dir, "scripts", "LINKER_wrapper.sh") + " "
        make_command += "make "
        os.system(make_command)
        # execute instrumented program
        os.system("./prog")
        # execute DiscoPoP analysis
        os.chdir(".discopop")
        os.system("discopop_explorer")
        os.chdir("..")
        # validate results
        self.validate_results(current_dir, src_dir)
        # clean environment
        os.system("make veryclean")

    def validate_results(self, test_dir, src_dir):
        """compare results to gold standard"""
        gold_standard_file = os.path.join(test_dir, "detection_result_dump.json")
        test_output_file = os.path.join(src_dir, ".discopop", "explorer", "detection_result_dump.json")
        # load both detection results
        with open(gold_standard_file, "r") as f:
            tmp_str = f.read()
        gold_standard: DetectionResult = jsonpickle.decode(tmp_str)

        with open(test_output_file, "r") as f:
            tmp_str = f.read()
        test_output: DetectionResult = jsonpickle.decode(tmp_str)

        # convert DoAllInfo objects to DoAllInfoForValidation objects to make use of custom __eq__
        converted_gold_standard = [DoAllInfoForValidation(elem) for elem in gold_standard.do_all]
        converted_test_output = [DoAllInfoForValidation(elem) for elem in test_output.patterns.do_all]

        # sort the lists
        converted_gold_standard = sorted(
            converted_gold_standard, key=lambda x: (x.dai.node_id, x.dai.start_line, x.dai.end_line)
        )
        converted_test_output = sorted(
            converted_test_output, key=lambda x: (x.dai.node_id, x.dai.start_line, x.dai.end_line)
        )

        # compare doall list elements
        self.assertListEqual(converted_gold_standard, converted_test_output)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.


import math
from typing import Dict, List, Tuple, Optional

from alive_progress import alive_bar  # type: ignore

from .PatternInfo import PatternInfo
from ..PEGraphX import FunctionNode, LoopNode, NodeID, PEGraphX, Node, EdgeType
from ..utils import classify_task_vars, get_child_loops
from ..variable import Variable

__loop_iterations: Dict[NodeID, int] = {}


class GDInfo(PatternInfo):
    """Class, that contains geometric decomposition detection result"""

    def __init__(self, pet: PEGraphX, node: Node, min_iter: int):
        """
        :param pet: PET graph
        :param node: node, where geometric decomposition was detected
        """
        PatternInfo.__init__(self, node)

        self.do_all_children, self.reduction_children = get_child_loops(pet, node)

        self.min_iter_number = min_iter
        mi_sqrt = math.sqrt(min_iter)
        wl = math.sqrt(self.get_workload(pet))
        nt = 1.1 * mi_sqrt + 0.0002 * wl - 0.0000002 * mi_sqrt * wl - 10

        if nt >= 1000:
            self.num_tasks = math.floor(nt / 100) * 100
        elif nt >= 100:
            self.num_tasks = math.floor(nt / 10) * 10
        elif nt < 0:
            self.num_tasks = 2
        else:
            self.num_tasks = math.floor(nt)

        self.pragma = "for (i = 0; i < num-tasks; i++) #pragma omp task"
        lp: List[Variable] = []
        fp, p, s, in_dep, out_dep, in_out_dep, r = classify_task_vars(pet, node, "GeometricDecomposition", [], [])
        fp.append(Variable("int", "i", "", sizeInByte=4))

        self.first_private = fp
        self.private = p
        self.last_private = lp
        self.shared = s
        self.reduction = r

    def __str__(self):
        return (
            f"Geometric decomposition at: {self.node_id}\n"
            f"Start line: {self.start_line}\n"
            f"End line: {self.end_line}\n"
            f"Do-All loops: {[n.id for n in self.do_all_children]}\n"
            f"Reduction loops: {[n.id for n in self.reduction_children]}\n"
            f"\tNumber of tasks: {self.num_tasks}\n"
            f"\tChunk limits: {self.min_iter_number}\n"
            f"\tpragma: {self.pragma}]\n"
            f"\tprivate: {[v.name for v in self.private]}\n"
            f"\tshared: {[v.name for v in self.shared]}\n"
            f"\tfirst private: {[v.name for v in self.first_private]}\n"
            f"\treduction: {[v for v in self.reduction]}\n"
            f"\tlast private: {[v.name for v in self.last_private]}"
        )


global_pet = None


def run_detection(pet: PEGraphX) -> List[GDInfo]:
    """Detects geometric decomposition

    :param pet: PET graph
    :return: List of detected pattern info
    """
    import tqdm  # type: ignore
    from multiprocessing import Pool

    global global_pet
    global_pet = pet

    result: List[GDInfo] = []
    global __loop_iterations
    __loop_iterations = {}
    nodes = pet.all_nodes(FunctionNode)

    param_list = [(node) for node in nodes]
    with Pool(initializer=__initialize_worker, initargs=(pet,)) as pool:
        tmp_result = list(tqdm.tqdm(pool.imap_unordered(__check_node, param_list), total=len(param_list)))
    for local_result in tmp_result:
        result += local_result
    print("GLOBAL RES: ", result)

    for pattern in result:
        pattern.get_workload(pet)

    return result


def __initialize_worker(pet):
    global global_pet
    global_pet = pet


def __check_node(param_tuple):
    global global_pet
    local_result = []
    node = param_tuple

    if __detect_geometric_decomposition(global_pet, node):
        node.geometric_decomposition = True
        test, min_iter = __test_chunk_limit(global_pet, node)
        if test and min_iter is not None:
            local_result.append(GDInfo(global_pet, node, min_iter))

    return local_result


def __test_chunk_limit(pet: PEGraphX, node: Node) -> Tuple[bool, Optional[int]]:
    """Tests, whether or not the node has inner loops with and none of them have 0 iterations

    :param pet: PET graph
    :param node: the node
    :return: true if node satisfies condition, min iteration number
    """
    min_iterations_count = None
    inner_loop_iter = {}

    children = pet.direct_children_or_called_nodes_of_type(node, LoopNode)

    for func_child in pet.direct_children_or_called_nodes_of_type(node, FunctionNode):
        children.extend(pet.direct_children_or_called_nodes_of_type(func_child, LoopNode))

    for child in children:
        inner_loop_iter[child.start_position()] = __iterations_count(pet, child)

    for k, v in inner_loop_iter.items():
        if min_iterations_count is None or v < min_iterations_count:
            min_iterations_count = v
    return (
        bool(inner_loop_iter) and (min_iterations_count is None or min_iterations_count > 0),
        min_iterations_count,
    )


def __iterations_count(pet: PEGraphX, node: LoopNode) -> int:
    """Counts the iterations in the specified node

    :param pet: PET graph
    :param node: the loop node
    :return: number of iterations
    """
    if not (node in __loop_iterations):
        loop_iter = node.loop_iterations
        parent_iter = __get_parent_iterations(pet, node)

        if loop_iter < parent_iter:
            __loop_iterations[node.id] = loop_iter
        elif loop_iter <= 0 or parent_iter <= 0:
            __loop_iterations[node.id] = 0
        else:
            __loop_iterations[node.id] = loop_iter // parent_iter

    return __loop_iterations[node.id]


def __get_parent_iterations(pet: PEGraphX, node: Node) -> int:
    """Calculates the number of iterations in parent of loop

    :param pet: PET graph
    :param node: current node
    :return: number of iterations
    """
    parent = pet.in_edges(node.id, [EdgeType.CHILD, EdgeType.CALLSNODE])

    max_iter = 1
    visited = []  # used to prevent looping
    while parent:
        node = pet.node_at(parent[0][0])
        # prevent looping
        if node in visited:
            break
        if isinstance(node, LoopNode):
            max_iter = max(1, node.loop_iterations)
            break
        visited.append(node)
        parent = pet.in_edges(node.id, [EdgeType.CHILD, EdgeType.CALLSNODE])

    return max_iter


def __detect_geometric_decomposition(pet: PEGraphX, root: Node) -> bool:
    """Detects geometric decomposition pattern

    :param pet: PET graph
    :param root: root node
    :return: true if GD pattern was discovered
    """
    for loop_child in pet.subtree_of_type(root, LoopNode):
        if not (loop_child.reduction or loop_child.do_all):
            return False

    for child in pet.direct_children_or_called_nodes_of_type(root, FunctionNode):
        for child2 in pet.direct_children_or_called_nodes_of_type(child, LoopNode):
            if not (child2.reduction or child2.do_all):
                return False

    return True



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Optional

from sympy import Function, Symbol, Integer, Expr  # type: ignore

from discopop_explorer.PEGraphX import NodeID
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.nodes.Workload import Workload


class FunctionRoot(Workload):
    name: str
    parallelizable_costs: Expr
    sequential_costs: Expr
    performance_model: CostModel

    def __init__(self, node_id: int, experiment, cu_id: Optional[NodeID], name: str):
        super().__init__(
            node_id,
            experiment,
            cu_id,
            sequential_workload=Integer(0),
            parallelizable_workload=Integer(0),
        )
        self.name = name
        self.device_id = experiment.get_system().get_host_device_id()
        function_name = "function" + "_" + str(self.node_id) + "_" + self.name
        self.parallelizable_costs = Symbol(function_name + "-parallelizable")
        self.sequential_costs = Symbol(function_name + "-sequential")
        self.performance_model = CostModel(
            self.parallelizable_costs,
            self.sequential_costs,
            identifier=function_name,
        )

    def get_plot_label(self) -> str:
        return self.name

    def get_cost_model(self, experiment, all_function_nodes, current_device) -> CostModel:
        """Model:
        Spawn overhead + children"""
        # todo this is only a dummy, not a finished model!
        function_name = "function" + "_" + str(self.node_id) + "_" + self.name
        # spawn_overhead = Symbol(function_name + "_spawn_overhead")
        # self.introduced_symbols.append(spawn_overhead)
        # model = Function(function_name)
        # model = spawn_overhead

        # todo: check if the costs of calling functions should be included into the models
        # self.performance_model = CostModel(Integer(0), Integer(0), identifier=function_name)

        # return CostModel(Integer(0), Integer(0), identifier=function_name)
        # instead of returning an empty model, return symbols for sequential and parallel workload.
        # These symbols can be substituted in order to evaluate for different versions of other functions

        cm = self.performance_model
        # substitute Expr(0) with Integer(0)
        cm.parallelizable_costs = cm.parallelizable_costs.subs({Expr(Integer(0)): Integer(0)})
        cm.sequential_costs = cm.sequential_costs.subs({Expr(Integer(0)): Integer(0)})

        return cm



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import copy
import json
from multiprocessing import Pool
import os
from typing import Dict, List, Optional, Set, Tuple, cast
import warnings

from sympy import Expr
import tqdm  # type: ignore
from discopop_explorer.PEGraphX import NodeID  # type: ignore
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments

from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
import random

from discopop_library.discopop_optimizer.optimization.evaluate import evaluate_configuration
from discopop_library.discopop_optimizer.optimization.validation import check_configuration_validity
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    get_in_options,
    get_out_mutex_edges,
    get_out_options,
    get_requirements,
)
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at
from discopop_library.result_classes.OptimizerOutputPattern import OptimizerOutputPattern


def perform_evolutionary_search(
    experiment: Experiment,
    available_decisions: Dict[FunctionRoot, List[List[int]]],
    arguments: OptimizerArguments,
    optimizer_dir: str,
) -> Optional[OptimizerOutputPattern]:
    if arguments.evolutionary is None:
        raise ValueError("Invalid arguments for evolutionary search: " + str(arguments.evolutionary))
    ### SETTINGS
    population_size = int(arguments.evolutionary[0])
    generations = int(arguments.evolutionary[1])
    selection_strength = 0.85  # 0.8 --> 80% of the population will be selected for the next generation
    crossovers = int(population_size / 10)
    mutations = int(population_size / 10)
    ### END SETTINGS

    population: List[List[int]] = __initialize(experiment, population_size, available_decisions, arguments)
    population, fitness, _ = __calculate_fitness(experiment, population, arguments)
    generation_counter = 0

    while generation_counter < generations:
        print("\nGeneration", generation_counter, "/", generations)
        population, fitness, _ = __calculate_fitness(experiment, population, arguments)
        if arguments.verbose:
            __print_population(experiment, population, fitness, arguments)
        population = __select(
            experiment,
            arguments,
            population,
            fitness,
            int(len(population) * selection_strength),
        )
        population = __fill_population(experiment, available_decisions, arguments, population, population_size)
        population = __crossover(experiment, arguments, population, crossovers)
        population = __mutate(experiment, arguments, population, crossovers)
        generation_counter += 1
    population, fitness, contexts = __calculate_fitness(experiment, population, arguments)
    if arguments.verbose:
        __print_population(experiment, population, fitness, arguments)
    return __dump_result(experiment, population, fitness, optimizer_dir, population_size, generations, contexts)


global_experiment = None
global_arguments = None
global_available_decisions = None
global_population = None


def __calculate_fitness(
    experiment: Experiment,
    population: List[List[int]],
    arguments: OptimizerArguments,
) -> Tuple[List[List[int]], List[int], List[ContextObject]]:
    """returning the population is necessary since the order of the population can change due to multiprocessing"""
    global global_experiment
    global global_arguments
    global_experiment = experiment
    global_arguments = arguments

    print("Calculating fitness...")
    param_list = [(element) for element in population]
    with Pool(
        initializer=__initialize_fitness_worker,
        initargs=(
            experiment,
            arguments,
        ),
    ) as pool:
        tmp_result = list(tqdm.tqdm(pool.imap_unordered(__get_score, param_list), total=len(param_list)))
    population = []
    fitness = []
    contexts = []
    for local_result in tmp_result:
        # remove invalid elements
        if local_result[1] == -1:
            continue

        population.append(local_result[0])
        fitness.append(local_result[1])
        contexts.append(local_result[2])
    return population, fitness, contexts


def __initialize_fitness_worker(
    experiment: Experiment,
    arguments: OptimizerArguments,
):
    global global_experiment
    global global_arguments
    global_experiment = experiment
    global_arguments = arguments


def __get_score(param_tuple) -> Tuple[List[int], int, ContextObject]:
    global global_experiment
    global global_arguments
    configuration = param_tuple
    try:
        _, score_expr, context = evaluate_configuration(
            cast(Experiment, global_experiment),
            configuration,
            cast(OptimizerArguments, global_arguments),
        )
        result = int(float(str(score_expr.evalf())))
    except ValueError:
        result = -1

    return configuration, result, context


def __print_population(
    experiment: Experiment,
    population: List[List[int]],
    fitness: List[int],
    arguments: OptimizerArguments,
):
    sorted_fitness = sorted(enumerate(fitness), key=lambda x: x[1], reverse=True)
    print("# POPULATION")
    for fitness_idx, fitness_value in sorted_fitness:
        element_with_mapping = []
        for entry in population[fitness_idx]:
            # find pattern id
            for pattern_id in experiment.suggestion_to_node_ids_dict:
                if entry in experiment.suggestion_to_node_ids_dict[pattern_id]:
                    element_with_mapping.append(
                        str(pattern_id) + "@" + str(data_at(experiment.optimization_graph, entry).device_id)
                    )
        print("#", element_with_mapping, "->", fitness_value)
    print("# AVG: ", int(sum(fitness) / len(fitness)))
    print()


def __initialize(
    experiment: Experiment,
    population_size: int,
    available_decisions: Dict[FunctionRoot, List[List[int]]],
    arguments: OptimizerArguments,
) -> List[List[int]]:
    return __fill_population(experiment, available_decisions, arguments, [], population_size)


def __initialize_fill_worker(
    experiment: Experiment,
    available_decisions: Dict[FunctionRoot, List[List[int]]],
    arguments: OptimizerArguments,
):
    global global_experiment
    global global_arguments
    global global_available_decisions
    global_experiment = experiment
    global_arguments = arguments
    global_available_decisions = available_decisions


def __parallel_get_random_configuration(param_tuple):
    global global_experiment
    global global_arguments
    global global_available_decisions
    return __get_random_configuration(global_experiment, global_available_decisions, global_arguments)


def __fill_population(
    experiment: Experiment,
    available_decisions: Dict[FunctionRoot, List[List[int]]],
    arguments: OptimizerArguments,
    population: List[List[int]],
    population_size: int,
) -> List[List[int]]:
    global global_experiment
    global global_arguments
    global global_available_decisions
    global_experiment = experiment
    global_arguments = arguments
    global_available_decisions = available_decisions  # type: ignore
    # select random candidates
    print("Filling the population...")
    param_list = [(None) for element in range(len(population), population_size)]
    with Pool(
        initializer=__initialize_fill_worker,
        initargs=(
            experiment,
            available_decisions,
            arguments,
        ),
    ) as pool:
        tmp_result = list(
            tqdm.tqdm(pool.imap_unordered(__parallel_get_random_configuration, param_list), total=len(param_list))
        )

    #    tmp_result = []
    #    for p in param_list:
    #        tmp_result.append(__parallel_get_random_configuration(p))

    for local_result in tmp_result:
        population.append(local_result)
    return population


def __select(
    experiment: Experiment,
    arguments: OptimizerArguments,
    population: List[List[int]],
    fitness: List[int],
    new_population_size: int,
):
    """Performs a fitness-proportionate Selection"""
    # get Sum of scores
    score_sum = 0.0
    for val in fitness:
        score_sum += val

    # get Probabilities for each element
    probabilityMap: Dict[Tuple[int, ...], float] = dict()
    for element_idx, element in enumerate(population):
        probabilityMap[tuple(element)] = 1 - (fitness[element_idx] / score_sum)

    # get weights list
    weights = [probabilityMap[tuple(elem)] for elem in population]

    # select population to preserve
    new_population = random.choices(population, weights=weights, k=new_population_size - 1)

    # always preserve the current best element
    for idx, fitness_value in sorted(enumerate(fitness), key=lambda x: x[1]):
        new_population.append(population[idx])
        break
    return new_population


def __crossover(
    experiment: Experiment,
    arguments: OptimizerArguments,
    population: List[List[int]],
    crossovers: int,
):
    global global_experiment
    global global_arguments
    global global_population
    global_experiment = experiment
    global_arguments = arguments
    global_population = population

    print("Calculating crossovers...")
    param_list = [(None) for element in range(0, crossovers)]
    with Pool(
        initializer=__initialize_crossover_worker,
        initargs=(experiment, arguments, population),
    ) as pool:
        tmp_result = list(tqdm.tqdm(pool.imap_unordered(__parallel_crossover, param_list), total=len(param_list)))
    for local_result in tmp_result:
        if local_result is None:
            continue
        if local_result[1] is not None:
            (old_element_1, old_element_2), (new_element_1, new_element_2) = local_result
            old_element_1, old_element_2 = local_result[1]
            if old_element_1 in population and old_element_2 in population and old_element_1 != old_element_2:
                population.remove(old_element_1)
                population.remove(old_element_2)
                population.append(new_element_1)
                population.append(new_element_2)
    return population


def __initialize_crossover_worker(experiment: Experiment, arguments: OptimizerArguments, population: List[List[int]]):
    global global_experiment
    global global_arguments
    global global_population
    global_experiment = experiment
    global_arguments = arguments
    global_population = population


def __parallel_crossover(param_tuple):
    global global_experiment
    global global_arguments
    global global_population

    for i in range(0, 1000):
        # select two random elements
        element_1 = random.choice(global_population)
        element_2 = random.choice(global_population)
        # select crossover point
        max_crossover_idx = min(len(element_1), len(element_2))
        crossover_idx = random.choice(range(0, max_crossover_idx))

        new_element_1 = element_1[:crossover_idx] + element_2[crossover_idx:]
        new_element_2 = element_2[:crossover_idx] + element_1[crossover_idx:]

        # validate elements
        if not check_configuration_validity(global_experiment, global_arguments, new_element_1):
            continue
        if not check_configuration_validity(global_experiment, global_arguments, new_element_2):
            continue
        return (element_1, element_2), (new_element_1, new_element_2)
    return None


def __mutate(
    experiment: Experiment,
    arguments: OptimizerArguments,
    population: List[List[int]],
    mutations: int,
) -> List[List[int]]:
    global global_experiment
    global global_arguments
    global global_population
    global_experiment = experiment
    global_arguments = arguments
    global_population = population

    print("Calculating mutations...")
    param_list = [(None) for element in range(0, mutations)]
    with Pool(
        initializer=__initialize_mutate_worker,
        initargs=(experiment, arguments, population),
    ) as pool:
        tmp_result = list(tqdm.tqdm(pool.imap_unordered(__parallel_mutate, param_list), total=len(param_list)))
    for local_result in tmp_result:
        if local_result is None:
            continue
        if local_result[0] in population and local_result[1] is not None:
            population.remove(local_result[0])
            population.append(local_result[1])
    return population


def __initialize_mutate_worker(experiment: Experiment, arguments: OptimizerArguments, population: List[List[int]]):
    global global_experiment
    global global_arguments
    global global_population
    global_experiment = experiment
    global_arguments = arguments
    global_population = population


def __parallel_mutate(param_tuple):
    global global_experiment
    global global_arguments
    global global_population

    for i in range(0, 1000):
        # select random mutation target from population
        mutation_target = random.choice(global_population)

        # select random mutation within the target
        mutation_index = random.choice(range(0, len(mutation_target)))

        # perform mutation if possible
        options = get_out_mutex_edges(global_experiment.optimization_graph, mutation_target[mutation_index])
        if len(options) > 0:
            index_mutant = random.choice(options)
            mutant = copy.deepcopy(mutation_target)
            mutant[mutation_index] = index_mutant

            # validate
            if not check_configuration_validity(global_experiment, global_arguments, mutant):
                continue
            return mutation_target, mutant
    return None


def __dump_result(
    experiment: Experiment,
    population: List[List[int]],
    fitness: List[int],
    optimizer_dir: str,
    population_size: int,
    generations: int,
    contexts: List[ContextObject],
) -> Optional[OptimizerOutputPattern]:
    # replace keys to allow dumping
    dumpable_dict = dict()
    for idx, key in enumerate(population):
        new_key = []
        for entry in key:
            # find pattern id
            for pattern_id in experiment.suggestion_to_node_ids_dict:
                if entry in experiment.suggestion_to_node_ids_dict[pattern_id]:
                    new_key.append(str(pattern_id) + "@" + str(data_at(experiment.optimization_graph, entry).device_id))
        dumpable_dict[str(new_key)] = str(fitness[idx])

    dump_path: str = os.path.join(
        optimizer_dir, "evolutionary_results_" + str(population_size) + "x" + str(generations) + ".json"
    )
    with open(dump_path, "w") as fp:
        json.dump(dumpable_dict, fp)

    # prepare dumping the best option
    for idx, fitness_value in sorted(enumerate(fitness), key=lambda x: x[1]):
        new_key_2 = []
        best_configuration = None
        for node_id in population[idx]:
            # find pattern id
            for pattern_id in experiment.suggestion_to_node_ids_dict:
                if node_id in experiment.suggestion_to_node_ids_dict[pattern_id]:
                    new_key_2.append(
                        str(pattern_id) + "@" + str(data_at(experiment.optimization_graph, node_id).device_id)
                    )
                    device_id = data_at(experiment.optimization_graph, node_id).device_id
                    if best_configuration is None:
                        best_configuration = OptimizerOutputPattern(
                            experiment.detection_result.pet.node_at(
                                cast(NodeID, data_at(experiment.optimization_graph, node_id).original_cu_id)
                            ),
                            [],
                            experiment.get_system().get_host_device_id(),
                        )
                    best_configuration.add_pattern(
                        pattern_id, device_id, experiment.get_system().get_device(device_id).get_device_type()
                    )
        if best_configuration is None:
            return None
        # collect data movement information
        for update in contexts[idx].necessary_updates:
            best_configuration.add_data_movement(update)
        # export results to file
        best_option_id_path: str = os.path.join(optimizer_dir, "evolutionary_pattern_id.txt")
        with open(best_option_id_path, "w+") as f:
            f.write(str(best_configuration.pattern_id))

        return best_configuration
    raise ValueError("No configuration found!")


def __get_random_configuration(
    experiment: Experiment,
    available_decisions: Dict[FunctionRoot, List[List[int]]],
    arguments: OptimizerArguments,
):
    while True:
        random_configuration: List[int] = []
        # fill configuration
        for function in available_decisions:
            excluded: Set[int] = set()
            requirements: Set[int] = set()

            for decision_list in available_decisions[function]:
                decision_set = set(decision_list)
                decision_set = decision_set - (decision_set & excluded)
                reduced_decision_set = decision_set.intersection(requirements)
                if len(reduced_decision_set) != 0:
                    random_decision = random.choice(list(reduced_decision_set))
                else:
                    random_decision = random.choice(list(decision_set))
                random_configuration.append(random_decision)
                requirements.update(get_requirements(experiment.optimization_graph, random_decision))
                excluded.update(get_out_mutex_edges(experiment.optimization_graph, random_decision))

        # validate configuration
        if check_configuration_validity(experiment, arguments, random_configuration):
            return random_configuration
